@@ -1,17 +1,21 @@
-#include <filezilla.h>
+#include "filezilla.h"
+#include "activity_logger_layer.h"
 #include "controlsocket.h"
 #include "directorycache.h"
 #include "engineprivate.h"
-#include "local_path.h"
 #include "lookup.h"
 #include "logging_private.h"
 #include "proxy.h"
 #include "servercapabilities.h"
-#include "sizeformatting_base.h"
+
+#include "../include/local_path.h"
+#include "../include/engine_options.h"
+#include "../include/sizeformatting_base.h"
 
 #include <libfilezilla/event_loop.hpp>
 #include <libfilezilla/iputils.hpp>
 #include <libfilezilla/local_filesys.hpp>
+#include <libfilezilla/rate_limited_layer.hpp>
 
 #include <string.h>
 
@@ -29,12 +33,13 @@
 	#endif
 #endif
 
-CControlSocket::CControlSocket(CFileZillaEnginePrivate & engine)
+CControlSocket::CControlSocket(CFileZillaEnginePrivate & engine, bool use_shm)
 	: event_handler(engine.event_loop_)
 	, engine_(engine)
 	, opLockManager_(engine.opLockManager_)
 	, logger_(engine.GetLogger())
 {
+	InitBufferPool(use_shm);
 }
 
 CControlSocket::~CControlSocket()
@@ -58,7 +63,7 @@ Command CControlSocket::GetCurrentCommandId() const
 		return operations_.back()->opId;
 	}
 
-	return engine_.GetCurrentCommandId();
+	return Command::none;
 }
 
 void CControlSocket::LogTransferResultMessage(int nErrorCode, CFileTransferOpData *pData)
@@ -136,18 +141,19 @@ int CControlSocket::ResetOperation(int nErrorCode)
 		nErrorCode = oldOperation->Reset(nErrorCode);
 	}
 	if (!operations_.empty()) {
-		int ret;
 		if (nErrorCode == FZ_REPLY_OK ||
 			nErrorCode == FZ_REPLY_ERROR ||
 			nErrorCode == FZ_REPLY_CRITICALERROR ||
 			nErrorCode == FZ_REPLY_ERROR_NOTFOUND)
 		{
-			ret = ParseSubcommandResult(nErrorCode, *oldOperation);
+			if (!oldOperation->topLevelOperation_) {
+				return ParseSubcommandResult(nErrorCode, std::move(oldOperation));
+			}
 		}
 		else {
-			ret = ResetOperation(nErrorCode);
+			oldOperation.reset();
+			return ResetOperation(nErrorCode);
 		}
-		return ret;
 	}
 
 	std::wstring prefix;
@@ -193,7 +199,7 @@ int CControlSocket::ResetOperation(int nErrorCode)
 		case Command::transfer:
 			{
 				auto & data = static_cast<CFileTransferOpData &>(*oldOperation);
-				if (!data.download_ && data.transferInitiated_) {
+				if (!data.download() && data.transferInitiated_) {
 					if (!currentServer_) {
 						log(logmsg::debug_warning, L"currentServer_ is empty");
 					}
@@ -210,18 +216,24 @@ int CControlSocket::ResetOperation(int nErrorCode)
 			}
 			break;
 		}
+
+		oldOperation.reset();
 	}
 
 	engine_.transfer_status_.Reset();
 
-	SetWait(false);
-
 	if (m_invalidateCurrentPath) {
 		currentPath_.clear();
 		m_invalidateCurrentPath = false;
 	}
 
-	return engine_.ResetOperation(nErrorCode);
+	if (operations_.empty()) {
+		SetWait(false);
+		return engine_.ResetOperation(nErrorCode);
+	}
+	else {
+		return SendNextCommand();
+	}
 }
 
 void CControlSocket::UpdateCache(COpData const &, CServerPath const& serverPath, std::wstring const& remoteFile, int64_t fileSize)
@@ -235,33 +247,28 @@ void CControlSocket::UpdateCache(COpData const &, CServerPath const& serverPath,
 int CControlSocket::DoClose(int nErrorCode)
 {
 	log(logmsg::debug_debug, L"CControlSocket::DoClose(%d)", nErrorCode);
-	if (m_closed) {
-		assert(operations_.empty());
-		return nErrorCode;
-	}
-
-	m_closed = true;
-
-	nErrorCode = ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED | nErrorCode);
-
-	currentServer_.clear();
-
-	return nErrorCode;
+	currentPath_.clear();
+	return ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED | nErrorCode);
 }
 
 std::wstring CControlSocket::ConvertDomainName(std::wstring const& domain)
 {
 #ifdef FZ_WINDOWS
-	int len = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), domain.size() + 1, nullptr, 0);
+	if (domain.size() > 1204 * 1024) {
+		return domain;
+	}
+
+	int const inlen = static_cast<int>(domain.size() + 1);
+	int len = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), inlen, nullptr, 0);
 	if (!len) {
 		log(logmsg::debug_warning, L"Could not convert domain name");
 		return domain;
 	}
 
 	wchar_t* output = new wchar_t[len];
-	int res = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), domain.size() + 1, output, len);
+	int res = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), inlen, output, len);
 	if (!res) {
-		delete [] output;
+		delete[] output;
 		log(logmsg::debug_warning, L"Could not convert domain name");
 		return domain;
 	}
@@ -303,39 +310,37 @@ CServer const& CControlSocket::GetCurrentServer() const
 	return currentServer_;
 }
 
-bool CControlSocket::ParsePwdReply(std::wstring reply, bool unquoted, CServerPath const& defaultPath)
+bool CControlSocket::ParsePwdReply(std::wstring reply, CServerPath const& defaultPath)
 {
-	if (!unquoted) {
-		size_t pos1 = reply.find('"');
-		size_t pos2 = reply.rfind('"');
-		// Due to searching the same character, pos1 is npos iff pos2 is npos
+	size_t pos1 = reply.find('"');
+	size_t pos2 = reply.rfind('"');
+	// Due to searching the same character, pos1 is npos iff pos2 is npos
 
-		if (pos1 == std::wstring::npos || pos1 >= pos2) {
-			pos1 = reply.find('\'');
-			pos2 = reply.rfind('\'');
+	if (pos1 == std::wstring::npos || pos1 >= pos2) {
+		pos1 = reply.find('\'');
+		pos2 = reply.rfind('\'');
 
-			if (pos1 != std::wstring::npos && pos1 < pos2) {
-				log(logmsg::debug_info, L"Broken server sending single-quoted path instead of double-quoted path.");
-			}
+		if (pos1 != std::wstring::npos && pos1 < pos2) {
+			log(logmsg::debug_info, L"Broken server sending single-quoted path instead of double-quoted path.");
 		}
-		if (pos1 == std::wstring::npos || pos1 >= pos2) {
-			log(logmsg::debug_info, L"Broken server, no quoted path found in pwd reply, trying first token as path");
-			pos1 = reply.find(' ');
-			if (pos1 != std::wstring::npos) {
-				reply = reply.substr(pos1 + 1);
-				pos2 = reply.find(' ');
-				if (pos2 != std::wstring::npos)
-					reply = reply.substr(0, pos2);
-			}
-			else {
-				reply.clear();
-			}
+	}
+	if (pos1 == std::wstring::npos || pos1 >= pos2) {
+		log(logmsg::debug_info, L"Broken server, no quoted path found in pwd reply, trying first token as path");
+		pos1 = reply.find(' ');
+		if (pos1 != std::wstring::npos) {
+			reply = reply.substr(pos1 + 1);
+			pos2 = reply.find(' ');
+			if (pos2 != std::wstring::npos)
+				reply = reply.substr(0, pos2);
 		}
 		else {
-			reply = reply.substr(pos1 + 1, pos2 - pos1 - 1);
-			fz::replace_substrings(reply, L"\"\"", L"\"");
+			reply.clear();
 		}
 	}
+	else {
+		reply = reply.substr(pos1 + 1, pos2 - pos1 - 1);
+		fz::replace_substrings(reply, L"\"\"", L"\"");
+	}
 
 	currentPath_.SetType(currentServer_.GetType());
 	if (reply.empty() || !currentPath_.SetPath(reply)) {
@@ -366,9 +371,11 @@ int CControlSocket::CheckOverwriteFile()
 	}
 
 	auto & data = static_cast<CFileTransferOpData &>(*operations_.back());
+	data.localFileSize_ = data.download() ? data.writer_factory_.size() : data.reader_factory_.size();
+	data.localFileTime_ = data.download() ? data.writer_factory_.mtime() : data.reader_factory_.mtime();
 
-	if (data.download_) {
-		if (fz::local_filesys::get_file_type(fz::to_native(data.localFile_), true) != fz::local_filesys::file) {
+	if (data.download()) {
+		if (data.localFileSize_ == fz::aio_base::nosize && data.localFileTime_.empty()) {
 			return FZ_REPLY_OK;
 		}
 	}
@@ -386,56 +393,79 @@ int CControlSocket::CheckOverwriteFile()
 	bool found = engine_.GetDirectoryCache().LookupFile(entry, currentServer_, remotePath, data.remoteFile_, dirDidExist, matchedCase);
 
 	// Ignore entries with wrong case
-	if (found && !matchedCase)
+	if (found && !matchedCase) {
 		found = false;
+	}
 
-	if (!data.download_) {
-		if (!found && data.remoteFileSize_ < 0 && data.fileTime_.empty()) {
+	if (found) {
+		if (data.remoteFileTime_.empty() && entry.has_date()) {
+			data.remoteFileTime_ = entry.time;
+		}
+	}
+
+	if (!data.download()) {
+		if (!found && data.remoteFileSize_ < 0 && data.remoteFileTime_.empty()) {
 			return FZ_REPLY_OK;
 		}
 	}
 
-	CFileExistsNotification *pNotification = new CFileExistsNotification;
 
-	pNotification->download = data.download_;
-	pNotification->localFile = data.localFile_;
-	pNotification->remoteFile = data.remoteFile_;
-	pNotification->remotePath = data.remotePath_;
-	pNotification->localSize = data.localFileSize_;
-	pNotification->remoteSize = data.remoteFileSize_;
-	pNotification->remoteTime = data.fileTime_;
-	pNotification->ascii = !data.transferSettings_.binary;
 
-	if (data.download_ && pNotification->localSize >= 0) {
-		pNotification->canResume = true;
+	auto notification = std::make_unique<CFileExistsNotification>();
+
+	notification->download = data.download();
+	notification->localFile = data.localName_;
+	notification->remoteFile = data.remoteFile_;
+	notification->remotePath = data.remotePath_;
+	notification->localSize = static_cast<int64_t>(data.localFileSize_);
+	notification->localTime = data.localFileTime_;
+	notification->remoteSize = data.remoteFileSize_;
+	notification->remoteTime = data.remoteFileTime_;
+
+	if (currentServer_.HasFeature(ProtocolFeature::DataTypeConcept)) {
+		notification->ascii = data.flags_ & ftp_transfer_flags::ascii;
 	}
-	else if (!data.download_ && pNotification->remoteSize >= 0) {
-		pNotification->canResume = true;
+
+	if (data.download() && notification->localSize >= 0) {
+		notification->canResume = true;
+	}
+	else if (!data.download() && notification->remoteSize >= 0) {
+		notification->canResume = true;
 	}
 	else {
-		pNotification->canResume = false;
+		notification->canResume = false;
 	}
 
-	pNotification->localTime = fz::local_filesys::get_modification_time(fz::to_native(data.localFile_));
-
-	if (found) {
-		if (pNotification->remoteTime.empty() && entry.has_date()) {
-			pNotification->remoteTime = entry.time;
-			data.fileTime_ = entry.time;
-		}
-	}
-
-	SendAsyncRequest(pNotification);
+	SendAsyncRequest(std::move(notification));
 
 	return FZ_REPLY_WOULDBLOCK;
 }
 
-CFileTransferOpData::CFileTransferOpData(wchar_t const* name, bool is_download, std::wstring const& local_file, std::wstring const& remote_file, CServerPath const& remote_path, CFileTransferCommand::t_transferSettings const& settings)
-	: COpData(Command::transfer, name)
-	, localFile_(local_file), remoteFile_(remote_file), remotePath_(remote_path)
-	, download_(is_download)
-	, transferSettings_(settings)
+SleepOpData::SleepOpData(CControlSocket & controlSocket, fz::duration const& delay)
+	: COpData(Command::sleep, L"SleepOpData")
+	, fz::event_handler(controlSocket.event_loop_)
+	, controlSocket_(controlSocket)
 {
+	add_timer(delay, true);
+	controlSocket_.SetWait(false);
+}
+
+void SleepOpData::operator()(fz::event_base const&)
+{
+	controlSocket_.ResetOperation(FZ_REPLY_OK);
+}
+
+CFileTransferOpData::CFileTransferOpData(wchar_t const* name, CFileTransferCommand const& cmd)
+	: COpData(Command::transfer, name)
+	, flags_(cmd.GetFlags())
+	, reader_factory_(cmd.GetReader())
+	, writer_factory_(cmd.GetWriter())
+	, localName_(reader_factory_ ? reader_factory_.name() : writer_factory_.name())
+	, remoteFile_(cmd.GetRemoteFile())
+	, remotePath_(cmd.GetRemotePath())
+{
+	localFileSize_ = download() ? writer_factory_.size() : reader_factory_.size();
+	localFileTime_ = download() ? writer_factory_.mtime() : reader_factory_.mtime();
 }
 
 std::wstring CControlSocket::ConvToLocal(char const* buffer, size_t len)
@@ -504,11 +534,11 @@ void CControlSocket::OnTimer(fz::timer_id)
 {
 	m_timer = 0; // It's a one-shot timer, no need to stop it
 
-	int const timeout = engine_.GetOptions().GetOptionVal(OPTION_TIMEOUT);
+	int const timeout = engine_.GetOptions().get_int(OPTION_TIMEOUT);
 	if (timeout > 0) {
 		fz::duration elapsed = fz::monotonic_clock::now() - m_lastActivity;
 
-		if ((operations_.empty() || !operations_.back()->waitForAsyncRequest) && !opLockManager_.Waiting(this)) {
+		if ((operations_.empty() || operations_.back()->async_request_state_ == async_request_state::none) && !opLockManager_.Waiting(this)) {
 			if (elapsed > fz::duration::from_seconds(timeout)) {
 				log(logmsg::error, fztranslate("Connection timed out after %d second of inactivity", "Connection timed out after %d seconds of inactivity", timeout), timeout);
 				DoClose(FZ_REPLY_TIMEOUT);
@@ -537,7 +567,7 @@ void CControlSocket::SetWait(bool wait)
 
 		m_lastActivity = fz::monotonic_clock::now();
 
-		int timeout = engine_.GetOptions().GetOptionVal(OPTION_TIMEOUT);
+		int timeout = engine_.GetOptions().get_int(OPTION_TIMEOUT);
 		if (!timeout) {
 			return;
 		}
@@ -561,7 +591,7 @@ int CControlSocket::SendNextCommand()
 
 	while (!operations_.empty()) {
 		auto & data = *operations_.back();
-		if (data.waitForAsyncRequest) {
+		if (data.async_request_state_ == async_request_state::waiting) {
 			log(logmsg::debug_info, L"Waiting for async request, ignoring SendNextCommand...");
 			return FZ_REPLY_WOULDBLOCK;
 		}
@@ -596,17 +626,21 @@ int CControlSocket::SendNextCommand()
 	return FZ_REPLY_OK;
 }
 
-int CControlSocket::ParseSubcommandResult(int prevResult, COpData const& opData)
+int CControlSocket::ParseSubcommandResult(int prevResult, std::unique_ptr<COpData> && previousOperation)
 {
 	if (operations_.empty()) {
 		log(logmsg::debug_warning, L"CControlSocket::ParseSubcommandResult(%d) called without active operation", prevResult);
+		previousOperation.reset();
 		ResetOperation(FZ_REPLY_ERROR);
 		return FZ_REPLY_ERROR;
 	}
 
 	auto & data = *operations_.back();
 	log(logmsg::debug_verbose, L"%s::SubcommandResult(%d) in state %d", data.name_, prevResult, data.opState);
-	int res = data.SubcommandResult(prevResult, opData);
+	int res = data.SubcommandResult(prevResult, *previousOperation);
+
+	previousOperation.reset();
+
 	if (res == FZ_REPLY_WOULDBLOCK) {
 		return FZ_REPLY_WOULDBLOCK;
 	}
@@ -630,10 +664,9 @@ void CControlSocket::OnObtainLock()
 	}
 }
 
-void CControlSocket::InvalidateCurrentWorkingDir(const CServerPath& path)
+void CControlSocket::InvalidateCurrentWorkingDir(CServerPath const& path)
 {
-	assert(!path.empty());
-	if (currentPath_.empty()) {
+	if (path.empty() || currentPath_.empty()) {
 		return;
 	}
 
@@ -647,29 +680,29 @@ void CControlSocket::InvalidateCurrentWorkingDir(const CServerPath& path)
 	}
 }
 
-fz::duration CControlSocket::GetTimezoneOffset() const
+fz::duration CControlSocket::GetInferredTimezoneOffset() const
 {
 	fz::duration ret;
 	if (currentServer_) {
 		int seconds = 0;
-		if (CServerCapabilities::GetCapability(currentServer_, timezone_offset, &seconds) == yes) {
+		if (CServerCapabilities::GetCapability(currentServer_, inferred_timezone_offset, &seconds) == yes) {
 			ret = fz::duration::from_seconds(seconds);
 		}
 	}
 	return ret;
 }
 
-void CControlSocket::SendAsyncRequest(CAsyncRequestNotification* pNotification)
+void CControlSocket::SendAsyncRequest(std::unique_ptr<CAsyncRequestNotification> && notification, bool wait)
 {
-	assert(pNotification);
-	assert(!operations_.empty());
-
-	pNotification->requestNumber = engine_.GetNextAsyncRequestNumber();
+	if (!notification || operations_.empty()) {
+		return;
+	}
+	notification->requestNumber = engine_.GetNextAsyncRequestNumber();
 
 	if (!operations_.empty()) {
-		operations_.back()->waitForAsyncRequest = true;
+		operations_.back()->async_request_state_ = wait ? async_request_state::waiting : async_request_state::parallel;
 	}
-	engine_.AddNotification(pNotification);
+	engine_.AddNotification(std::move(notification));
 }
 
 // ------------------
@@ -686,11 +719,6 @@ CRealControlSocket::~CRealControlSocket()
 	ResetSocket();
 }
 
-bool CRealControlSocket::Connected() const
-{
-	return socket_ ? socket_->is_connected() : false;
-}
-
 int CRealControlSocket::Send(unsigned char const* buffer, unsigned int len)
 {
 	if (!active_layer_) {
@@ -715,7 +743,7 @@ int CRealControlSocket::Send(unsigned char const* buffer, unsigned int len)
 		}
 
 		if (written) {
-			SetActive(CFileZillaEngine::send);
+			SetAlive();
 		}
 
 		if (static_cast<unsigned int>(written) < len) {
@@ -816,7 +844,7 @@ int CRealControlSocket::OnSend()
 		}
 
 		if (written) {
-			SetActive(CFileZillaEngine::send);
+			RecordActivity(activity_logger::send, written);
 
 			send_buffer_.consume(static_cast<size_t>(written));
 		}
@@ -837,29 +865,24 @@ void CRealControlSocket::OnSocketError(int error)
 	DoClose();
 }
 
-int CRealControlSocket::DoConnect(std::wstring const& host, unsigned int port)
+void CRealControlSocket::CreateSocket(std::wstring const& host)
 {
-	SetWait(true);
-
-	if (currentServer_.GetEncodingType() == ENCODING_CUSTOM) {
-		log(logmsg::debug_info, L"Using custom encoding: %s", currentServer_.GetCustomEncoding());
-	}
-
 	ResetSocket();
 	socket_ = std::make_unique<fz::socket>(engine_.GetThreadPool(), nullptr);
-	ratelimit_layer_ = std::make_unique<CRatelimitLayer>(this, *socket_, engine_.GetRateLimiter());
+	activity_logger_layer_ = std::make_unique<activity_logger_layer>(nullptr, *socket_, engine_.activity_logger_);
+	ratelimit_layer_ = std::make_unique<fz::rate_limited_layer>(nullptr, *activity_logger_layer_, &engine_.GetRateLimiter());
 	active_layer_ = ratelimit_layer_.get();
 
-	const int proxy_type = engine_.GetOptions().GetOptionVal(OPTION_PROXY_TYPE);
+	const int proxy_type = engine_.GetOptions().get_int(OPTION_PROXY_TYPE);
 	if (proxy_type > static_cast<int>(ProxyType::NONE) && proxy_type < static_cast<int>(ProxyType::count) && !currentServer_.GetBypassProxy()) {
 		log(logmsg::status, _("Connecting to %s through %s proxy"), currentServer_.Format(ServerFormat::with_optional_port), CProxySocket::Name(static_cast<ProxyType>(proxy_type)));
 
-		fz::native_string proxy_host = fz::to_native(engine_.GetOptions().GetOption(OPTION_PROXY_HOST));
+		fz::native_string proxy_host = fz::to_native(engine_.GetOptions().get_string(OPTION_PROXY_HOST));
 
-		proxy_layer_ = std::make_unique<CProxySocket>(this, *active_layer_, this, static_cast<ProxyType>(proxy_type),
-			proxy_host, engine_.GetOptions().GetOptionVal(OPTION_PROXY_PORT),
-			engine_.GetOptions().GetOption(OPTION_PROXY_USER),
-			engine_.GetOptions().GetOption(OPTION_PROXY_PASS));
+		proxy_layer_ = std::make_unique<CProxySocket>(nullptr, *active_layer_, this, static_cast<ProxyType>(proxy_type),
+			proxy_host, engine_.GetOptions().get_int(OPTION_PROXY_PORT),
+			engine_.GetOptions().get_string(OPTION_PROXY_USER),
+			engine_.GetOptions().get_string(OPTION_PROXY_PASS));
 		active_layer_ = proxy_layer_.get();
 
 		if (fz::get_address_type(proxy_host) == fz::address_type::unknown) {
@@ -872,13 +895,25 @@ int CRealControlSocket::DoConnect(std::wstring const& host, unsigned int port)
 		}
 	}
 
-	m_closed = false;
+	SetSocketBufferSizes();
+}
 
+int CRealControlSocket::DoConnect(std::wstring const& host, unsigned int port)
+{
+	SetWait(true);
+
+	if (currentServer_.GetEncodingType() == ENCODING_CUSTOM) {
+		log(logmsg::debug_info, L"Using custom encoding: %s", currentServer_.GetCustomEncoding());
+	}
+
+	CreateSocket(host);
+
+	active_layer_->set_event_handler(this);
 	int res = active_layer_->connect(fz::to_native(ConvertDomainName(host)), port);
 
 	if (res) {
 		log(logmsg::error, _("Could not connect to server: %s"), fz::socket_error_description(res));
-		return FZ_REPLY_DISCONNECTED | FZ_REPLY_ERROR; 
+		return FZ_REPLY_DISCONNECTED | FZ_REPLY_ERROR;
 	}
 
 	return FZ_REPLY_WOULDBLOCK;
@@ -899,6 +934,7 @@ void CRealControlSocket::ResetSocket()
 	// Destroy in reverse order
 	proxy_layer_.reset();
 	ratelimit_layer_.reset();
+	activity_logger_layer_.reset();
 	socket_.reset();
 
 	send_buffer_.clear();
@@ -906,7 +942,9 @@ void CRealControlSocket::ResetSocket()
 
 bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNotification)
 {
-	assert(pFileExistsNotification);
+	if (!pFileExistsNotification) {
+		return false;
+	}
 
 	if (operations_.empty() || operations_.back()->opId != Command::transfer) {
 		log(logmsg::debug_info, L"SetFileExistsAction: No or invalid operation in progress, ignoring request reply %f", pFileExistsNotification->GetRequestID());
@@ -931,12 +969,12 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 			SendNextCommand();
 		}
 		else {
-			if (data.download_) {
+			if (data.download()) {
 				std::wstring filename = data.remotePath_.FormatFilename(data.remoteFile_);
 				log(logmsg::status, _("Skipping download of %s"), filename);
 			}
 			else {
-				log(logmsg::status, _("Skipping upload of %s"), data.localFile_);
+				log(logmsg::status, _("Skipping upload of %s"), data.localName_);
 			}
 			ResetOperation(FZ_REPLY_OK);
 		}
@@ -948,12 +986,12 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 			SendNextCommand();
 		}
 		else {
-			if (data.download_) {
+			if (data.download()) {
 				std::wstring filename = data.remotePath_.FormatFilename(data.remoteFile_);
 				log(logmsg::status, _("Skipping download of %s"), filename);
 			}
 			else {
-				log(logmsg::status, _("Skipping upload of %s"), data.localFile_);
+				log(logmsg::status, _("Skipping upload of %s"), data.localName_);
 			}
 			ResetOperation(FZ_REPLY_OK);
 		}
@@ -974,54 +1012,39 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 			SendNextCommand();
 		}
 		else {
-			if (data.download_) {
+			if (data.download()) {
 				auto const filename = data.remotePath_.FormatFilename(data.remoteFile_);
 				log(logmsg::status, _("Skipping download of %s"), filename);
 			}
 			else {
-				log(logmsg::status, _("Skipping upload of %s"), data.localFile_);
+				log(logmsg::status, _("Skipping upload of %s"), data.localName_);
 			}
 			ResetOperation(FZ_REPLY_OK);
 		}
 		break;
 	case CFileExistsNotification::resume:
-		if (data.download_ && data.localFileSize_ >= 0) {
+		if (data.download() && data.localFileSize_ != fz::aio_base::nosize) {
 			data.resume_ = true;
 		}
-		else if (!data.download_ && data.remoteFileSize_ >= 0) {
+		else if (!data.download() && data.remoteFileSize_ >= 0) {
 			data.resume_ = true;
 		}
 		SendNextCommand();
 		break;
 	case CFileExistsNotification::rename:
-		if (data.download_) {
+		if (data.download()) {
 			{
-				std::wstring tmp;
-				CLocalPath l(data.localFile_, &tmp);
-				if (l.empty() || tmp.empty()) {
-					ResetOperation(FZ_REPLY_INTERNALERROR);
-					return false;
-				}
-				if (!l.ChangePath(pFileExistsNotification->newName)) {
-					ResetOperation(FZ_REPLY_INTERNALERROR);
-					return false;
-				}
-				if (!l.HasParent() || !l.MakeParent(&tmp)) {
+				if (!pFileExistsNotification->new_writer_factory_) {
 					ResetOperation(FZ_REPLY_INTERNALERROR);
 					return false;
 				}
 
-				data.localFile_ = l.GetPath() + tmp;
+				data.writer_factory_ = std::move(pFileExistsNotification->new_writer_factory_);
+				data.localName_ = data.writer_factory_.name();
 			}
 
-			int64_t size;
-			bool isLink;
-			if (fz::local_filesys::get_file_info(fz::to_native(data.localFile_), isLink, &size, nullptr, nullptr) == fz::local_filesys::file) {
-				data.localFileSize_ = size;
-			}
-			else {
-				data.localFileSize_ = -1;
-			}
+			data.localFileSize_ = data.writer_factory_.size();
+			data.localFileTime_ = data.writer_factory_.mtime();
 
 			if (CheckOverwriteFile() == FZ_REPLY_OK) {
 				SendNextCommand();
@@ -1029,8 +1052,8 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 		}
 		else {
 			data.remoteFile_ = pFileExistsNotification->newName;
-			data.fileTime_ = fz::datetime();
 			data.remoteFileSize_ = -1;
+			data.remoteFileTime_ = fz::datetime();
 
 			CDirentry entry;
 			bool dir_did_exist;
@@ -1040,7 +1063,7 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 			{
 				data.remoteFileSize_ = entry.size;
 				if (entry.has_date()) {
-					data.fileTime_ = entry.time;
+					data.remoteFileTime_ = entry.time;
 				}
 
 				if (CheckOverwriteFile() != FZ_REPLY_OK) {
@@ -1052,12 +1075,12 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 		}
 		break;
 	case CFileExistsNotification::skip:
-		if (data.download_) {
+		if (data.download()) {
 			std::wstring filename = data.remotePath_.FormatFilename(data.remoteFile_);
 			log(logmsg::status, _("Skipping download of %s"), filename);
 		}
 		else {
-			log(logmsg::status, _("Skipping upload of %s"), data.localFile_);
+			log(logmsg::status, _("Skipping upload of %s"), data.localName_);
 		}
 		ResetOperation(FZ_REPLY_OK);
 		break;
@@ -1070,25 +1093,6 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 	return true;
 }
 
-void CControlSocket::CreateLocalDir(std::wstring const& local_file)
-{
-	std::wstring file;
-	CLocalPath local_path(local_file, &file);
-	if (local_path.empty() || !local_path.HasParent()) {
-		return;
-	}
-
-	CLocalPath last_successful;
-	local_path.Create(&last_successful);
-
-	if (!last_successful.empty()) {
-		// Send out notification
-		CLocalDirCreatedNotification *n = new CLocalDirCreatedNotification;
-		n->dir = last_successful;
-		engine_.AddNotification(n);
-	}
-}
-
 void CControlSocket::List(CServerPath const&, std::wstring const&, int)
 {
 	Push(std::make_unique<CNotSupportedOpData>());
@@ -1109,7 +1113,7 @@ void CControlSocket::RemoveDir(CServerPath const&, std::wstring const&)
 	Push(std::make_unique<CNotSupportedOpData>());
 }
 
-void CControlSocket::Mkdir(CServerPath const&)
+void CControlSocket::Mkdir(CServerPath const&, transfer_flags const&)
 {
 	Push(std::make_unique<CNotSupportedOpData>());
 }
@@ -1124,41 +1128,6 @@ void CControlSocket::Chmod(CChmodCommand const&)
 	Push(std::make_unique<CNotSupportedOpData>());
 }
 
-void CControlSocket::operator()(fz::event_base const& ev)
-{
-	fz::dispatch<fz::timer_event, CObtainLockEvent>(ev, this,
-		&CControlSocket::OnTimer,
-		&CControlSocket::OnObtainLock);
-}
-
-void CControlSocket::SetActive(CFileZillaEngine::_direction direction)
-{
-	SetAlive();
-	engine_.SetActive(direction);
-}
-
-void CControlSocket::SendDirectoryListingNotification(CServerPath const& path, bool failed)
-{
-	if (!currentServer_) {
-		return;
-	}
-
-	engine_.AddNotification(new CDirectoryListingNotification(path, operations_.size() == 1 && operations_.back()->opId == Command::list, failed));
-}
-
-void CControlSocket::CallSetAsyncRequestReply(CAsyncRequestNotification *pNotification)
-{
-	if (operations_.empty() || !operations_.back()->waitForAsyncRequest) {
-		log(logmsg::debug_info, L"Not waiting for request reply, ignoring request reply %d", pNotification->GetRequestID());
-		return;
-	}
-
-	operations_.back()->waitForAsyncRequest = false;
-
-	SetAlive();
-	SetAsyncRequestReply(pNotification);
-}
-
 void CControlSocket::Lookup(CServerPath const& path, std::wstring const& file, CDirentry * entry)
 {
 	Push(std::make_unique<LookupOpData>(*this, path, file, entry));
@@ -1169,6 +1138,87 @@ void CControlSocket::Lookup(CServerPath const& path, std::vector<std::wstring> c
 	Push(std::make_unique<LookupManyOpData>(*this, path, files));
 }
 
+void CControlSocket::operator()(fz::event_base const& ev)
+{
+	fz::dispatch<fz::timer_event, CObtainLockEvent>(ev, this,
+		&CControlSocket::OnTimer,
+		&CControlSocket::OnObtainLock);
+}
+
+void CControlSocket::RecordActivity(activity_logger::_direction direction, uint64_t amount)
+{
+	SetAlive();
+	engine_.activity_logger_.record(direction, amount);
+}
+
+void CControlSocket::SendDirectoryListingNotification(CServerPath const& path, bool failed)
+{
+	if (!currentServer_) {
+		return;
+	}
+
+	engine_.AddNotification(std::make_unique<CDirectoryListingNotification>(path, operations_.size() == 1 && operations_.back()->opId == Command::list, failed));
+}
+
+void CControlSocket::CallSetAsyncRequestReply(CAsyncRequestNotification *pNotification)
+{
+	if (operations_.empty() || operations_.back()->async_request_state_ == async_request_state::none) {
+		log(logmsg::debug_info, L"Not waiting for request reply, ignoring request reply %d", pNotification->GetRequestID());
+		return;
+	}
+
+	operations_.back()->async_request_state_ = async_request_state::none;
+
+	SetAlive();
+	SetAsyncRequestReply(pNotification);
+}
+
+void CControlSocket::Sleep(fz::duration const& delay)
+{
+	Push(std::make_unique<SleepOpData>(*this, delay));
+}
+
+bool CControlSocket::InitBufferPool(bool use_shm)
+{
+	if (!buffer_pool_) {
+		buffer_pool_.emplace(logger_, 8, 0, use_shm);
+	}
+	return *buffer_pool_;
+}
+
+std::unique_ptr<fz::writer_base> CControlSocket::OpenWriter(fz::writer_factory_holder & factory, uint64_t resumeOffset, bool withProgress)
+{
+	if (!factory || !buffer_pool_) {
+		return {};
+	}
+
+	auto file_writer = dynamic_cast<fz::file_writer_factory*>(&*factory);
+	if (file_writer) {
+		std::wstring tmp;
+		CLocalPath local_path(file_writer->name(), &tmp);
+		if (local_path.HasParent()) {
+			fz::native_string last_created;
+			fz::mkdir(fz::to_native(local_path.GetPath()), true, fz::mkdir_permissions::normal, &last_created);
+			if (!last_created.empty()) {
+				// Send out notification
+				auto n = std::make_unique<CLocalDirCreatedNotification>();
+				if (n->dir.SetPath(fz::to_wstring(last_created))) {
+					engine_.AddNotification(std::move(n));
+				}
+			}
+		}
+	}
+
+	fz::writer_base::progress_cb_t status_update;
+	if (withProgress) {
+		status_update = [&s = engine_.transfer_status_](fz::writer_base const*, uint64_t written) {
+			s.SetMadeProgress();
+			s.Update(written);
+		};
+	}
+	return factory->open(*buffer_pool_, resumeOffset, status_update, max_buffer_count());
+}
+
 int64_t CalculateNextChunkSize(int64_t remaining, int64_t lastChunkSize, fz::duration const& lastChunkDuration, int64_t minChunkSize, int64_t multiple, int64_t partCount, int64_t maxPartCount, int64_t maxChunkSize)
 {
 	if (remaining <= 0) {
@@ -1197,7 +1247,7 @@ int64_t CalculateNextChunkSize(int64_t remaining, int64_t lastChunkSize, fz::dur
 
 	if (multiple) {
 		// Round up
-		int modulus = newChunkSize % multiple;
+		auto modulus = newChunkSize % multiple;
 		if (modulus) {
 			newChunkSize += multiple - modulus;
 		}
@@ -1213,3 +1263,13 @@ int64_t CalculateNextChunkSize(int64_t remaining, int64_t lastChunkSize, fz::dur
 
 	return newChunkSize;
 }
+
+int64_t CalculateNextChunkSize(int64_t remaining, int64_t lastChunkSize, fz::monotonic_clock const& lastChunkStart, int64_t minChunkSize, int64_t multiple, int64_t partCount, int64_t maxPartCount, int64_t maxChunkSize)
+{
+	return CalculateNextChunkSize(remaining, lastChunkSize, fz::monotonic_clock::now() - lastChunkStart, minChunkSize, multiple, partCount, maxPartCount, maxChunkSize);
+}
+
+size_t CControlSocket::max_buffer_count() const
+{
+	return buffer_pool_->buffer_count();
+}
