@@ -1,29 +1,35 @@
-#include <filezilla.h>
+#include "filezilla.h"
 #include "controlsocket.h"
 #include "directorycache.h"
 #include "engineprivate.h"
+#if ENABLE_FTP
 #include "ftp/ftpcontrolsocket.h"
+#endif
 #include "http/httpcontrolsocket.h"
 #include "logging_private.h"
 #include "pathcache.h"
-#include "ratelimiter.h"
+#if ENABLE_SFTP
 #include "sftp/sftpcontrolsocket.h"
+#endif
 #if ENABLE_STORJ
 #include "storj/storjcontrolsocket.h"
 #endif
 
+#include "../include/engine_options.h"
+
 #include <libfilezilla/event_loop.hpp>
 
 #include <algorithm>
 
 fz::mutex CFileZillaEnginePrivate::global_mutex_{false};
 std::vector<CFileZillaEnginePrivate*> CFileZillaEnginePrivate::m_engineList;
-std::atomic_int CFileZillaEnginePrivate::m_activeStatus[2] = {{0}, {0}};
 std::list<CFileZillaEnginePrivate::t_failedLogins> CFileZillaEnginePrivate::m_failedLogins;
 
 struct invalid_current_working_dir_event_type{};
 typedef fz::simple_event<invalid_current_working_dir_event_type, CServer, CServerPath> CInvalidateCurrentWorkingDirEvent;
 
+struct command_event_type{};
+typedef fz::simple_event<command_event_type> CCommandEvent;
 
 namespace {
 unsigned int get_next_engine_id()
@@ -33,14 +39,15 @@ unsigned int get_next_engine_id()
 }
 }
 
-CFileZillaEnginePrivate::CFileZillaEnginePrivate(CFileZillaEngineContext& context, CFileZillaEngine& parent, EngineNotificationHandler& notificationHandler)
+CFileZillaEnginePrivate::CFileZillaEnginePrivate(CFileZillaEngineContext& context, CFileZillaEngine& parent, std::function<void(CFileZillaEngine*)> const& notification_cb)
 	: event_handler(context.GetEventLoop())
 	, transfer_status_(*this)
 	, opLockManager_(context.GetOpLockManager())
-	, notification_handler_(notificationHandler)
+	, activity_logger_(context.GetActivityLogger())
+	, notification_cb_(notification_cb)
 	, m_engine_id(get_next_engine_id())
-	, m_options(context.GetOptions())
-	, m_rateLimiter(context.GetRateLimiter())
+	, options_(context.GetOptions())
+	, rate_limiter_(context.GetRateLimiter())
 	, directory_cache_(context.GetDirectoryCache())
 	, path_cache_(context.GetPathCache())
 	, parent_(parent)
@@ -61,44 +68,73 @@ CFileZillaEnginePrivate::CFileZillaEnginePrivate(CFileZillaEngineContext& contex
 		queue_logs_ = queue_logs;
 	}
 
-	RegisterOption(OPTION_LOGGING_SHOW_DETAILED_LOGS);
-	RegisterOption(OPTION_LOGGING_DEBUGLEVEL);
-	RegisterOption(OPTION_LOGGING_RAWLISTING);
+	options_.watch(OPTION_LOGGING_SHOW_DETAILED_LOGS, this);
+	options_.watch(OPTION_LOGGING_DEBUGLEVEL, this);
+	options_.watch(OPTION_LOGGING_RAWLISTING, this);
 }
 
 bool CFileZillaEnginePrivate::ShouldQueueLogsFromOptions() const
 {
 	return
-		m_options.GetOptionVal(OPTION_LOGGING_RAWLISTING) == 0 &&
-		m_options.GetOptionVal(OPTION_LOGGING_DEBUGLEVEL) == 0 &&
-		m_options.GetOptionVal(OPTION_LOGGING_SHOW_DETAILED_LOGS) == 0;
+		options_.get_int(OPTION_LOGGING_RAWLISTING) == 0 &&
+		options_.get_int(OPTION_LOGGING_DEBUGLEVEL) == 0 &&
+		options_.get_int(OPTION_LOGGING_SHOW_DETAILED_LOGS) == 0;
+}
+
+namespace {
+// Removes item from a random access container,
+// filling the hole by moving the last item.
+template<typename C, typename V>
+void erase_unordered(C & container, V const& v) noexcept
+{
+	for (size_t i = 0; i < container.size(); ++i) {
+		if (container[i] == v) {
+			if (i + 1 < container.size()) {
+				container[i] = std::move(container.back());
+			}
+			container.pop_back();
+			return;
+		}
+	}
+}
 }
 
 CFileZillaEnginePrivate::~CFileZillaEnginePrivate()
 {
-	remove_handler();
-	UnregisterAllOptions();
+	shutdown();
+}
 
-	m_maySendNotificationEvent = false;
+void CFileZillaEnginePrivate::shutdown()
+{
+	options_.unwatch_all(this);
+	remove_handler();
+
+	{
+		fz::scoped_lock lock(notification_mutex_);
+		m_maySendNotificationEvent = false;
+
+		auto cb = std::move(notification_cb_);
+		notification_cb_ = nullptr;
+		// Must not hold mutex when destroying cb, or we'll deadlock
+		lock.unlock();
+	}
 
 	controlSocket_.reset();
-	m_pCurrentCommand.reset();
+	currentCommand_.reset();
 
-	// Delete notification list
-	for (auto & notification : m_NotificationList) {
-		delete notification;
+	{
+		fz::scoped_lock lock(notification_mutex_);
+		// Delete notification list
+		for (auto & notification : m_NotificationList) {
+			delete notification;
+		}
+		m_NotificationList.clear();
 	}
 
 	// Remove ourself from the engine list
 	{
 		fz::scoped_lock lock(global_mutex_);
-		m_engineList.erase(std::remove(m_engineList.begin(), m_engineList.end(), this), m_engineList.end());
-		for (auto iter = m_engineList.begin(); iter != m_engineList.end(); ++iter) {
-			if (*iter == this) {
-				m_engineList.erase(iter);
-				break;
-			}
-		}
+		erase_unordered(m_engineList, this);
 	}
 }
 
@@ -117,95 +153,72 @@ void CFileZillaEnginePrivate::OnEngineEvent(EngineNotificationType type)
 bool CFileZillaEnginePrivate::IsBusy() const
 {
 	fz::scoped_lock lock(mutex_);
-	return m_pCurrentCommand != nullptr;
+	return currentCommand_ != nullptr;
 }
 
 bool CFileZillaEnginePrivate::IsConnected() const
 {
 	fz::scoped_lock lock(mutex_);
-	if (!controlSocket_) {
-		return false;
-	}
-
-	// FIXME: This is unsafe!
-	return controlSocket_->Connected();
+	return controlSocket_ != nullptr;
 }
 
-const CCommand *CFileZillaEnginePrivate::GetCurrentCommand() const
+void CFileZillaEnginePrivate::AddNotification(fz::scoped_lock&, std::unique_ptr<CNotification> && notification)
 {
-	fz::scoped_lock lock(mutex_);
-	return m_pCurrentCommand.get();
-}
-
-Command CFileZillaEnginePrivate::GetCurrentCommandId() const
-{
-	fz::scoped_lock lock(mutex_);
-	if (!m_pCurrentCommand) {
-		return Command::none;
+	if (notification) {
+		m_NotificationList.push_back(notification.release());
 	}
-	else {
-		return m_pCurrentCommand->GetId();
-	}
-}
 
-void CFileZillaEnginePrivate::AddNotification(fz::scoped_lock& lock, CNotification *pNotification)
-{
-	m_NotificationList.push_back(pNotification);
-
-	if (m_maySendNotificationEvent) {
+	if (m_maySendNotificationEvent && notification_cb_) {
 		m_maySendNotificationEvent = false;
-		lock.unlock();
-		notification_handler_.OnEngineEvent(&parent_);
+		notification_cb_(&parent_);
 	}
 }
 
-void CFileZillaEnginePrivate::AddNotification(CNotification *pNotification)
+void CFileZillaEnginePrivate::AddNotification(std::unique_ptr<CNotification> && notification)
 {
 	fz::scoped_lock lock(notification_mutex_);
-	AddNotification(lock, pNotification);
+	AddNotification(lock, std::move(notification));
 }
 
-void CFileZillaEnginePrivate::AddLogNotification(CLogmsgNotification *pNotification)
+void CFileZillaEnginePrivate::AddLogNotification(std::unique_ptr<CLogmsgNotification> && notification)
 {
 	fz::scoped_lock lock(notification_mutex_);
 
-	if (pNotification->msgType == logmsg::error) {
+	if (notification->msgType == logmsg::error) {
 		queue_logs_ = false;
 
 		m_NotificationList.insert(m_NotificationList.end(), queued_logs_.begin(), queued_logs_.end());
 		queued_logs_.clear();
-		AddNotification(lock, pNotification);
+		AddNotification(lock, std::move(notification));
 	}
-	else if (pNotification->msgType == logmsg::status) {
+	else if (notification->msgType == logmsg::status) {
 		ClearQueuedLogs(lock, false);
-		AddNotification(lock, pNotification);
+		AddNotification(lock, std::move(notification));
 	}
 	else if (!queue_logs_) {
-		AddNotification(lock, pNotification);
+		AddNotification(lock, std::move(notification));
 	}
 	else {
-		queued_logs_.push_back(pNotification);
+		queued_logs_.push_back(notification.release());
 	}
 }
 
 void CFileZillaEnginePrivate::SendQueuedLogs(bool reset_flag)
 {
-	{
-		fz::scoped_lock lock(notification_mutex_);
-		m_NotificationList.insert(m_NotificationList.end(), queued_logs_.begin(), queued_logs_.end());
-		queued_logs_.clear();
+	fz::scoped_lock lock(notification_mutex_);
+	m_NotificationList.insert(m_NotificationList.end(), queued_logs_.begin(), queued_logs_.end());
+	queued_logs_.clear();
 
-		if (reset_flag) {
-			queue_logs_ = ShouldQueueLogsFromOptions();
-		}
-
-		if (!m_maySendNotificationEvent || m_NotificationList.empty()) {
-			return;
-		}
-		m_maySendNotificationEvent = false;
+	if (reset_flag) {
+		queue_logs_ = ShouldQueueLogsFromOptions();
 	}
 
-	notification_handler_.OnEngineEvent(&parent_);
+	if (!m_maySendNotificationEvent || m_NotificationList.empty() || !notification_cb_) {
+		return;
+	}
+	m_maySendNotificationEvent = false;
+
+	notification_cb_(&parent_);
 }
 
 void CFileZillaEnginePrivate::ClearQueuedLogs(fz::scoped_lock&, bool reset_flag)
@@ -231,22 +244,26 @@ int CFileZillaEnginePrivate::ResetOperation(int nErrorCode)
 	fz::scoped_lock lock(mutex_);
 	logger_->log(logmsg::debug_debug, L"CFileZillaEnginePrivate::ResetOperation(%d)", nErrorCode);
 
-	if (m_pCurrentCommand) {
+	if (currentCommand_) {
 		if ((nErrorCode & FZ_REPLY_NOTSUPPORTED) == FZ_REPLY_NOTSUPPORTED) {
 			logger_->log(logmsg::error, _("Command not supported by this protocol"));
 		}
 
-		if (m_pCurrentCommand->GetId() == Command::connect) {
+		if (currentCommand_->GetId() == Command::connect) {
+			if (m_retryTimer) {
+				return FZ_REPLY_WOULDBLOCK;
+			}
+
 			if (!(nErrorCode & ~(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED | FZ_REPLY_TIMEOUT | FZ_REPLY_CRITICALERROR | FZ_REPLY_PASSWORDFAILED)) &&
 				nErrorCode & (FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED))
 			{
-				CConnectCommand const& connectCommand = static_cast<CConnectCommand const&>(*m_pCurrentCommand.get());
+				CConnectCommand const& connectCommand = static_cast<CConnectCommand const&>(*currentCommand_.get());
 
 				RegisterFailedLoginAttempt(connectCommand.GetServer(), (nErrorCode & FZ_REPLY_CRITICALERROR) == FZ_REPLY_CRITICALERROR);
 
 				if ((nErrorCode & FZ_REPLY_CRITICALERROR) != FZ_REPLY_CRITICALERROR) {
 					++m_retryCount;
-					if (m_retryCount < m_options.GetOptionVal(OPTION_RECONNECTCOUNT) && connectCommand.RetryConnecting()) {
+					if (m_retryCount < options_.get_int(OPTION_RECONNECTCOUNT) && connectCommand.RetryConnecting()) {
 						fz::duration delay = GetRemainingReconnectDelay(connectCommand.GetServer());
 						if (!delay) {
 							delay = fz::duration::from_seconds(1);
@@ -260,25 +277,9 @@ int CFileZillaEnginePrivate::ResetOperation(int nErrorCode)
 			}
 		}
 
-		if (!m_bIsInCommand) {
-			COperationNotification *notification = new COperationNotification();
-			notification->nReplyCode = nErrorCode;
-			notification->commandId = m_pCurrentCommand->GetId();
-			AddNotification(notification);
-		}
-		else {
-			m_nControlSocketError |= nErrorCode;
-		}
+		AddNotification(std::make_unique<COperationNotification>(nErrorCode, currentCommand_->GetId()));
 
-		m_pCurrentCommand.reset();
-	}
-	else if (nErrorCode & FZ_REPLY_DISCONNECTED) {
-		if (!m_bIsInCommand) {
-			COperationNotification *notification = new COperationNotification();
-			notification->nReplyCode = nErrorCode;
-			notification->commandId = Command::none;
-			AddNotification(notification);
-		}
+		currentCommand_.reset();
 	}
 
 	if (nErrorCode != FZ_REPLY_OK) {
@@ -293,8 +294,7 @@ int CFileZillaEnginePrivate::ResetOperation(int nErrorCode)
 
 unsigned int CFileZillaEnginePrivate::GetNextAsyncRequestNumber()
 {
-	fz::scoped_lock lock(notification_mutex_);
-	return ++m_asyncRequestCounter;
+	return ++asyncRequestCounter_;
 }
 
 // Command handlers
@@ -306,12 +306,6 @@ int CFileZillaEnginePrivate::Connect(CConnectCommand const& command)
 
 	m_retryCount = 0;
 
-	// Need to delete before setting m_pCurrentCommand.
-	// The destructor can call CFileZillaEnginePrivate::ResetOperation
-	// which would delete m_pCurrentCommand
-	controlSocket_.reset();
-	m_nControlSocketError = 0;
-
 	auto const& server = command.GetServer();
 	if (server.GetPort() != CServer::GetDefaultPort(server.GetProtocol())) {
 		ServerProtocol protocol = CServer::GetProtocolFromPort(server.GetPort(), true);
@@ -359,32 +353,29 @@ int CFileZillaEnginePrivate::List(CListCommand const& command)
 						server.GetProtocol() == SWIFT || server.GetProtocol() == GOOGLE_CLOUD ||
 						server.GetProtocol() == GOOGLE_DRIVE || server.GetProtocol() == DROPBOX ||
 						server.GetProtocol() == ONEDRIVE || server.GetProtocol() == B2 ||
-						server.GetProtocol() == BOX) {
+						server.GetProtocol() == BOX || server.GetProtocol() == RACKSPACE || server.GetProtocol() == STORJ_GRANT) {
 					path = command.GetPath();
 					path.ChangePath(command.GetSubDir());
 				}
 			}
 			if (!path.empty()) {
-				CDirectoryListing *pListing = new CDirectoryListing;
+				CDirectoryListing listing;
 				bool is_outdated = false;
-				bool found = directory_cache_.Lookup(*pListing, server, path, true, is_outdated);
+				bool found = directory_cache_.Lookup(listing, server, path, true, is_outdated);
 				if (found && !is_outdated) {
-					if (pListing->get_unsure_flags()) {
+					if (listing.get_unsure_flags()) {
 						flags |= LIST_FLAG_REFRESH;
 					}
 					else {
 						if (!avoid) {
-							CDirectoryListingNotification *pNotification = new CDirectoryListingNotification(pListing->path, true);
-							AddNotification(pNotification);
+							AddNotification(std::make_unique<CDirectoryListingNotification>(listing.path, true));
 						}
-						delete pListing;
 						return FZ_REPLY_OK;
 					}
 				}
 				if (is_outdated) {
 					flags |= LIST_FLAG_REFRESH;
 				}
-				delete pListing;
 			}
 		}
 	}
@@ -395,7 +386,7 @@ int CFileZillaEnginePrivate::List(CListCommand const& command)
 
 int CFileZillaEnginePrivate::FileTransfer(CFileTransferCommand const& command)
 {
-	controlSocket_->FileTransfer(command.GetLocalFile(), command.GetRemotePath(), command.GetRemoteFile(), command.Download(), command.GetTransferSettings());
+	controlSocket_->FileTransfer(command);
 	return FZ_REPLY_CONTINUE;
 }
 
@@ -429,7 +420,7 @@ int CFileZillaEnginePrivate::RemoveDir(CRemoveDirCommand const& command)
 
 int CFileZillaEnginePrivate::Mkdir(CMkdirCommand const& command)
 {
-	controlSocket_->Mkdir(command.GetPath());
+	controlSocket_->Mkdir(command.GetPath(), {});
 	return FZ_REPLY_CONTINUE;
 }
 
@@ -451,7 +442,7 @@ void CFileZillaEnginePrivate::RegisterFailedLoginAttempt(const CServer& server,
 	std::list<t_failedLogins>::iterator iter = m_failedLogins.begin();
 	while (iter != m_failedLogins.end()) {
 		fz::duration const span = fz::monotonic_clock::now() - iter->time;
-		if (span.get_seconds() >= m_options.GetOptionVal(OPTION_RECONNECTDELAY) ||
+		if (span.get_seconds() >= options_.get_int(OPTION_RECONNECTDELAY) ||
 			iter->server.SameResource(server) || (!critical && (iter->server.GetHost() == server.GetHost() && iter->server.GetPort() == server.GetPort())))
 		{
 			std::list<t_failedLogins>::iterator prev = iter;
@@ -476,7 +467,7 @@ fz::duration CFileZillaEnginePrivate::GetRemainingReconnectDelay(CServer const&
 	std::list<t_failedLogins>::iterator iter = m_failedLogins.begin();
 	while (iter != m_failedLogins.end()) {
 		fz::duration const span = fz::monotonic_clock::now() - iter->time;
-		fz::duration const delay = fz::duration::from_seconds(m_options.GetOptionVal(OPTION_RECONNECTDELAY));
+		fz::duration const delay = fz::duration::from_seconds(options_.get_int(OPTION_RECONNECTDELAY));
 		if (span >= delay) {
 			std::list<t_failedLogins>::iterator prev = iter;
 			++iter;
@@ -501,21 +492,18 @@ void CFileZillaEnginePrivate::OnTimer(fz::timer_id)
 	if (!m_retryTimer) {
 		return;
 	}
-	m_retryTimer = 0;
 
-	if (!m_pCurrentCommand || m_pCurrentCommand->GetId() != Command::connect) {
+	if (!currentCommand_ || currentCommand_->GetId() != Command::connect) {
+		m_retryTimer = 0;
 		logger_->log(logmsg::debug_warning, L"CFileZillaEnginePrivate::OnTimer called without pending Command::connect");
 		return;
 	}
 
 	controlSocket_.reset();
+	m_retryTimer = 0;
 
 	int res = ContinueConnect();
-	res |= m_nControlSocketError;
-	m_nControlSocketError = 0;
-
 	if (res == FZ_REPLY_CONTINUE) {
-		assert(controlSocket_);
 		controlSocket_->SendNextCommand();
 	}
 	else if (res != FZ_REPLY_WOULDBLOCK) {
@@ -527,12 +515,12 @@ int CFileZillaEnginePrivate::ContinueConnect()
 {
 	fz::scoped_lock lock(mutex_);
 
-	if (!m_pCurrentCommand || m_pCurrentCommand->GetId() != Command::connect) {
+	if (!currentCommand_ || currentCommand_->GetId() != Command::connect) {
 		logger_->log(logmsg::debug_warning, L"CFileZillaEnginePrivate::ContinueConnect called without pending Command::connect");
 		return ResetOperation(FZ_REPLY_INTERNALERROR);
 	}
 
-	const CConnectCommand *pConnectCommand = static_cast<CConnectCommand *>(m_pCurrentCommand.get());
+	const CConnectCommand *pConnectCommand = static_cast<CConnectCommand *>(currentCommand_.get());
 	const CServer& server = pConnectCommand->GetServer();
 	fz::duration const& delay = GetRemainingReconnectDelay(server);
 	if (delay) {
@@ -544,21 +532,26 @@ int CFileZillaEnginePrivate::ContinueConnect()
 
 	switch (server.GetProtocol())
 	{
+#if ENABLE_FTP
 	case FTP:
 	case FTPS:
 	case FTPES:
 	case INSECURE_FTP:
 		controlSocket_ = std::make_unique<CFtpControlSocket>(*this);
 		break;
+#endif
+#if ENABLE_SFTP
 	case SFTP:
 		controlSocket_ = std::make_unique<CSftpControlSocket>(*this);
 		break;
+#endif
 	case HTTP:
 	case HTTPS:
 		controlSocket_ = std::make_unique<CHttpControlSocket>(*this);
 		break;
 #if ENABLE_STORJ
 	case STORJ:
+	case STORJ_GRANT:
 		controlSocket_ = std::make_unique<CStorjControlSocket>(*this);
 		break;
 #endif
@@ -593,14 +586,14 @@ void CFileZillaEnginePrivate::InvalidateCurrentWorkingDirs(const CServerPath& pa
 		// May happen during destruction
 		return;
 	}
-	
+
 	fz::scoped_lock lock(global_mutex_);
 	for (auto & engine : m_engineList) {
 		if (!engine || engine == this) {
 			continue;
 		}
 
-		engine->send_event(new CInvalidateCurrentWorkingDirEvent(ownServer, path));
+		engine->send_event<CInvalidateCurrentWorkingDirEvent>(ownServer, path);
 	}
 }
 
@@ -608,22 +601,19 @@ void CFileZillaEnginePrivate::operator()(fz::event_base const& ev)
 {
 	fz::scoped_lock lock(mutex_);
 
-	fz::dispatch<CFileZillaEngineEvent, CCommandEvent, CAsyncRequestReplyEvent, fz::timer_event, CInvalidateCurrentWorkingDirEvent>(ev, this,
+	fz::dispatch<CFileZillaEngineEvent, CCommandEvent, CAsyncRequestReplyEvent, fz::timer_event, CInvalidateCurrentWorkingDirEvent, options_changed_event>(ev, this,
 		&CFileZillaEnginePrivate::OnEngineEvent,
 		&CFileZillaEnginePrivate::OnCommandEvent,
 		&CFileZillaEnginePrivate::OnSetAsyncRequestReplyEvent,
 		&CFileZillaEnginePrivate::OnTimer,
-		&CFileZillaEnginePrivate::OnInvalidateCurrentWorkingDir
+		&CFileZillaEnginePrivate::OnInvalidateCurrentWorkingDir,
+		&CFileZillaEnginePrivate::OnOptionsChanged
 		);
 }
 
 int CFileZillaEnginePrivate::CheckCommandPreconditions(CCommand const& command, bool checkBusy)
 {
-	if (!command.valid()) {
-		logger_->log(logmsg::debug_warning, L"Command not valid");
-		return FZ_REPLY_SYNTAXERROR;
-	}
-	else if (checkBusy && IsBusy()) {
+	if (checkBusy && IsBusy()) {
 		return FZ_REPLY_BUSY;
 	}
 	else if (command.GetId() != Command::connect && command.GetId() != Command::disconnect && !IsConnected()) {
@@ -639,8 +629,8 @@ void CFileZillaEnginePrivate::OnCommandEvent()
 {
 	fz::scoped_lock lock(mutex_);
 
-	if (m_pCurrentCommand) {
-		CCommand & command = *m_pCurrentCommand;
+	if (currentCommand_) {
+		CCommand & command = *currentCommand_;
 		Command id = command.GetId();
 
 		int res = CheckCommandPreconditions(command, false);
@@ -695,17 +685,17 @@ void CFileZillaEnginePrivate::OnCommandEvent()
 			}
 		}
 
-		if (id != Command::disconnect) {
-			res |= m_nControlSocketError;
-		}
-		else if (res & FZ_REPLY_DISCONNECTED) {
+		if (id == Command::disconnect && (res & FZ_REPLY_DISCONNECTED)) {
 			res = FZ_REPLY_OK;
 		}
-		m_nControlSocketError = 0;
 
 		if (res == FZ_REPLY_CONTINUE) {
-			assert(controlSocket_);
-			controlSocket_->SendNextCommand();
+			if (controlSocket_) {
+				controlSocket_->SendNextCommand();
+			}
+			else {
+				ResetOperation(FZ_REPLY_INTERNALERROR);
+			}
 		}
 		else if (res != FZ_REPLY_WOULDBLOCK) {
 			ResetOperation(res);
@@ -721,20 +711,15 @@ void CFileZillaEnginePrivate::DoCancel()
 	}
 
 	if (m_retryTimer) {
-		assert(m_pCurrentCommand && m_pCurrentCommand->GetId() == Command::connect);
-
 		controlSocket_.reset();
 
-		m_pCurrentCommand.reset();
+		currentCommand_.reset();
 
 		stop_timer(m_retryTimer);
 		m_retryTimer = 0;
 
 		logger_->log(logmsg::error, _("Connection attempt interrupted by user"));
-		COperationNotification *notification = new COperationNotification();
-		notification->nReplyCode = FZ_REPLY_DISCONNECTED | FZ_REPLY_CANCELED;
-		notification->commandId = Command::connect;
-		AddNotification(notification);
+		AddNotification(std::make_unique<COperationNotification>(FZ_REPLY_DISCONNECTED | FZ_REPLY_CANCELED, Command::connect));
 
 		ClearQueuedLogs(true);
 	}
@@ -748,36 +733,26 @@ void CFileZillaEnginePrivate::DoCancel()
 	}
 }
 
-bool CFileZillaEnginePrivate::CheckAsyncRequestReplyPreconditions(std::unique_ptr<CAsyncRequestNotification> const& reply)
-{
-	if (!reply) {
-		return false;
-	}
-	if (!IsBusy()) {
-		return false;
-	}
-
-	bool match;
-	{
-		fz::scoped_lock l(notification_mutex_);
-		match = reply->requestNumber == m_asyncRequestCounter;
-	}
-
-	return match && controlSocket_;
-}
-
 void CFileZillaEnginePrivate::OnSetAsyncRequestReplyEvent(std::unique_ptr<CAsyncRequestNotification> const& reply)
 {
 	fz::scoped_lock lock(mutex_);
-	if (!CheckAsyncRequestReplyPreconditions(reply)) {
+	if (!controlSocket_) {
+		return;
+	}
+	if (!IsPendingAsyncRequestReply(reply)) {
 		return;
 	}
 
 	controlSocket_->CallSetAsyncRequestReply(reply.get());
 }
 
-int CFileZillaEnginePrivate::Execute(const CCommand &command)
+int CFileZillaEnginePrivate::Execute(CCommand const& command)
 {
+	if (!command.valid()) {
+		logger_->log(logmsg::debug_warning, L"Command not valid");
+		return FZ_REPLY_SYNTAXERROR;
+	}
+
 	fz::scoped_lock lock(mutex_);
 
 	int res = CheckCommandPreconditions(command, true);
@@ -785,7 +760,7 @@ int CFileZillaEnginePrivate::Execute(const CCommand &command)
 		return res;
 	}
 
-	m_pCurrentCommand.reset(command.Clone());
+	currentCommand_.reset(command.Clone());
 	send_event<CCommandEvent>();
 
 	return FZ_REPLY_WOULDBLOCK;
@@ -808,7 +783,7 @@ std::unique_ptr<CNotification> CFileZillaEnginePrivate::GetNextNotification()
 bool CFileZillaEnginePrivate::SetAsyncRequestReply(std::unique_ptr<CAsyncRequestNotification> && pNotification)
 {
 	fz::scoped_lock lock(mutex_);
-	if (!CheckAsyncRequestReplyPreconditions(pNotification)) {
+	if (!IsPendingAsyncRequestReply(pNotification)) {
 		return false;
 	}
 
@@ -827,28 +802,7 @@ bool CFileZillaEnginePrivate::IsPendingAsyncRequestReply(std::unique_ptr<CAsyncR
 		return false;
 	}
 
-	fz::scoped_lock lock(notification_mutex_);
-	return pNotification->requestNumber == m_asyncRequestCounter;
-}
-
-void CFileZillaEnginePrivate::SetActive(int direction)
-{
-	int const old_status = m_activeStatus[direction].fetch_or(0x1);
-	if (!old_status) {
-		AddNotification(new CActiveNotification(direction));
-	}
-}
-
-bool CFileZillaEnginePrivate::IsActive(CFileZillaEngine::_direction direction)
-{
-	int const old = m_activeStatus[direction].exchange(0x2);
-	if (!(old & 0x1)) {
-		// Race: We might lose updates between the first exchange and this assignment.
-		// It is harmless though
-		m_activeStatus[direction] = 0;
-		return false;
-	}
-	return true;
+	return pNotification->requestNumber == asyncRequestCounter_;
 }
 
 CTransferStatus CFileZillaEnginePrivate::GetTransferStatus(bool &changed)
@@ -865,7 +819,9 @@ int CFileZillaEnginePrivate::CacheLookup(const CServerPath& path, CDirectoryList
 		return FZ_REPLY_ERROR;
 	}
 
-	assert(controlSocket_->GetCurrentServer());
+	if (!controlSocket_->GetCurrentServer()) {
+		return FZ_REPLY_INTERNALERROR;
+	}
 
 	bool is_outdated = false;
 	if (!directory_cache_.Lookup(listing, controlSocket_->GetCurrentServer(), path, true, is_outdated)) {
@@ -886,7 +842,7 @@ int CFileZillaEnginePrivate::Cancel()
 	return FZ_REPLY_WOULDBLOCK;
 }
 
-void CFileZillaEnginePrivate::OnOptionsChanged(changed_options_t const&)
+void CFileZillaEnginePrivate::OnOptionsChanged(watched_options const&)
 {
 	bool queue_logs = ShouldQueueLogsFromOptions();
 	if (queue_logs) {
@@ -917,7 +873,7 @@ void CTransferStatusManager::Reset()
 		send_state_ = 0;
 	}
 
-	engine_.AddNotification(new CTransferStatusNotification());
+	engine_.AddNotification(std::make_unique<CTransferStatusNotification>());
 }
 
 void CTransferStatusManager::Init(int64_t totalSize, int64_t startOffset, bool list)
@@ -929,6 +885,7 @@ void CTransferStatusManager::Init(int64_t totalSize, int64_t startOffset, bool l
 
 	status_ = CTransferStatus(totalSize, startOffset, list);
 	currentOffset_ = 0;
+	made_progress_ = false;
 }
 
 void CTransferStatusManager::SetStartTime()
@@ -943,17 +900,12 @@ void CTransferStatusManager::SetStartTime()
 
 void CTransferStatusManager::SetMadeProgress()
 {
-	fz::scoped_lock lock(mutex_);
-	if (!status_) {
-		return;
-	}
-
-	status_.madeProgress = true;
+	made_progress_ = true;
 }
 
 void CTransferStatusManager::Update(int64_t transferredBytes)
 {
-	CNotification* notification = nullptr;
+	std::unique_ptr<CNotification> notification;
 
 	{
 		int64_t oldOffset = currentOffset_.fetch_add(transferredBytes);
@@ -965,14 +917,15 @@ void CTransferStatusManager::Update(int64_t transferredBytes)
 
 			if (!send_state_) {
 				status_.currentOffset += currentOffset_.exchange(0);
-				notification = new CTransferStatusNotification(status_);
+				status_.madeProgress = made_progress_;
+				notification = std::make_unique<CTransferStatusNotification>(status_);
 			}
 			send_state_ = 2;
 		}
 	}
 
 	if (notification) {
-		engine_.AddNotification(notification);
+		engine_.AddNotification(std::move(notification));
 	}
 }
 
@@ -985,6 +938,7 @@ CTransferStatus CTransferStatusManager::Get(bool &changed)
 	}
 	else {
 		status_.currentOffset += currentOffset_.exchange(0);
+		status_.madeProgress = made_progress_;
 		if (send_state_ == 2) {
 			changed = true;
 			send_state_ = 1;
