@@ -1,4 +1,4 @@
-#include <filezilla.h>
+#include "filezilla.h"
 
 #include <libfilezilla/format.hpp>
 #include <libfilezilla/uri.hpp>
@@ -10,6 +10,7 @@ struct t_protocolInfo
 	ServerProtocol const protocol;
 	std::wstring const prefix;
 	bool alwaysShowPrefix;
+	bool parse_from_prefix;
 	unsigned int defaultPort;
 	bool const translateable;
 	char const* const name;
@@ -17,37 +18,46 @@ struct t_protocolInfo
 };
 
 static const t_protocolInfo protocolInfos[] = {
-	{ FTP,             L"ftp",      false, 21, true,  fztranslate_mark("FTP - File Transfer Protocol with optional encryption"), L"" },
-	{ SFTP,            L"sftp",     true,  22, false, "SFTP - SSH File Transfer Protocol",                                       L"" },
-	{ HTTP,            L"http",     true,  80, false, "HTTP - Hypertext Transfer Protocol",                                      L"" },
-	{ HTTPS,           L"https",    true, 443, true,  fztranslate_mark("HTTPS - HTTP over TLS"),                                 L"" },
-	{ FTPS,            L"ftps",     true, 990, true,  fztranslate_mark("FTPS - FTP over implicit TLS"),                          L"" },
-	{ FTPES,           L"ftpes",    true,  21, true,  fztranslate_mark("FTPES - FTP over explicit TLS"),                         L"" },
-	{ INSECURE_FTP,    L"ftp",      false, 21, true,  fztranslate_mark("FTP - Insecure File Transfer Protocol"),                 L"" },
-	{ S3,              L"s3",       true, 443, false, "S3 - Amazon Simple Storage Service",                                      L"" },
-	{ STORJ,           L"storj",    true, 443, true,  fztranslate_mark("Storj - Decentralized Cloud Storage"),                   L"" },
-	{ WEBDAV,          L"webdav",   true, 443, true,  fztranslate_mark("WebDAV using HTTPS"),                                    L"https" },
-	{ AZURE_FILE,      L"azfile",   true, 443, false, "Microsoft Azure File Storage Service",                                    L"https" },
-	{ AZURE_BLOB,      L"azblob",   true, 443, false, "Microsoft Azure Blob Storage Service",                                    L"https" },
-	{ SWIFT,           L"swift",    true, 443, false, "OpenStack Swift",                                                         L"https" },
-	{ GOOGLE_CLOUD,    L"google",   true, 443, false, "Google Cloud Storage",                                                    L"https" },
-	{ GOOGLE_DRIVE,    L"gdrive",   true, 443, false, "Google Drive",                                                            L"https" },
-	{ DROPBOX,         L"dropbox",  true, 443, false, "Dropbox",                                                                 L"https" },
-	{ ONEDRIVE,        L"onedrive", true, 443, false, "Microsoft OneDrive",                                                      L"https" },
-	{ B2,              L"b2",       true, 443, false, "Backblaze B2",                                                            L"https" },
-	{ BOX,             L"box",      true, 443, false, "Box",                                                                     L"https" },
-	{ INSECURE_WEBDAV, L"webdav",   true,  80, true,  fztranslate_mark("WebDAV using HTTP (insecure)"),                          L"http" },
-	{ UNKNOWN,         L"",         false, 21, false, "",                                                                        L"" }
+	{ FTP,             L"ftp",       false, true,    21, true,  fztranslate_mark("FTP - File Transfer Protocol with optional encryption"), L"" },
+	{ SFTP,            L"sftp",      true,  true,    22, false, "SFTP - SSH File Transfer Protocol",                                       L"" },
+	{ HTTP,            L"http",      true,  true,    80, false, "HTTP - Hypertext Transfer Protocol",                                      L"" },
+	{ HTTPS,           L"https",     true,  true,   443, true,  fztranslate_mark("HTTPS - HTTP over TLS"),                                 L"" },
+	{ FTPS,            L"ftps",      true,  true,   990, true,  fztranslate_mark("FTPS - FTP over implicit TLS"),                          L"" },
+	{ FTPES,           L"ftpes",     true,  true,    21, true,  fztranslate_mark("FTPES - FTP over explicit TLS"),                         L"" },
+	{ INSECURE_FTP,    L"ftp",       false, false,   21, true,  fztranslate_mark("FTP - Insecure File Transfer Protocol"),                 L"" },
+	{ S3,              L"s3",        true,  true,   443, false, "S3 - Amazon Simple Storage Service",                                      L"" },
+	{ STORJ,           L"storj",     true,  false, 7777, true,  fztranslate_mark("Storj (using legacy API key)"),                          L"" },
+	{ WEBDAV,          L"webdav",    true,  true,   443, true,  fztranslate_mark("WebDAV using HTTPS"),                                    L"https" },
+	{ AZURE_FILE,      L"azfile",    true,  true,   443, false, "Microsoft Azure File Storage Service",                                    L"https" },
+	{ AZURE_BLOB,      L"azblob",    true,  true,   443, false, "Microsoft Azure Blob Storage Service",                                    L"https" },
+	{ SWIFT,           L"swift",     true,  true,   443, false, "OpenStack Swift",                                                         L"https" },
+	{ GOOGLE_CLOUD,    L"google",    true,  true,   443, false, "Google Cloud Storage",                                                    L"https" },
+	{ GOOGLE_DRIVE,    L"gdrive",    true,  true,   443, false, "Google Drive",                                                            L"https" },
+	{ DROPBOX,         L"dropbox",   true,  true,   443, false, "Dropbox",                                                                 L"https" },
+	{ ONEDRIVE,        L"onedrive",  true,  true,   443, false, "Microsoft OneDrive",                                                      L"https" },
+	{ B2,              L"b2",        true,  true,   443, false, "Backblaze B2",                                                            L"https" },
+	{ BOX,             L"box",       true,  true,   443, false, "Box",                                                                     L"https" },
+	{ INSECURE_WEBDAV, L"webdav",    true,  true,    80, true,  fztranslate_mark("WebDAV using HTTP (insecure)"),                          L"http" },
+	{ RACKSPACE,       L"rackspace", true,  true,   443, false, "Rackspace Cloud Storage",                                                 L"https" },
+	{ STORJ_GRANT,     L"storj",     true,  true,  7777, true,  fztranslate_mark("Storj - Decentralized Cloud Storage"),                   L"" },
+	{ UNKNOWN,         L"",          false, false,   21, false, "",                                                                        L"" }
 };
 
 static std::vector<ServerProtocol> const defaultProtocols = {
+#if ENABLE_FTP
 	FTP,
+#endif
+#if ENABLE_SFTP
 	SFTP,
+#endif
+#if ENABLE_FTP
+	FTP,
 	FTPS,
 	FTPES,
 	INSECURE_FTP,
+#endif
 #if ENABLE_STORJ
-	STORJ,
+	STORJ_GRANT,
 #endif
 };
 
@@ -228,7 +238,6 @@ bool CServer::operator<(const CServer &op) const
 		return false;
 	}
 
-
 	// Do not compare number of allowed multiple connections
 
 	return false;
@@ -268,7 +277,11 @@ void CServer::SetProtocol(ServerProtocol serverProtocol)
 	m_protocol = serverProtocol;
 
 	// Clear out parameters not supported by the current protocol
-	std::map<std::string, std::wstring> oldParams;
+	if (!ProtocolHasUser(serverProtocol)) {
+		m_user.clear();
+	}
+
+	std::map<std::string, std::wstring, std::less<>> oldParams;
 	std::swap(extraParameters_, oldParams);
 	for (auto const& param : oldParams) {
 		SetExtraParameter(param.first, param.second);
@@ -355,15 +368,20 @@ std::wstring CServer::Format(ServerFormat formatType, Credentials const& credent
 		return server;
 	}
 
-	if (m_port != GetDefaultPort(m_protocol)) {
+	if (m_port != GetDefaultPort(m_protocol) || formatType == ServerFormat::with_port) {
 		server += fz::sprintf(L":%d", m_port);
 	}
 
-	if (formatType == ServerFormat::with_optional_port) {
+	if (formatType == ServerFormat::with_optional_port || formatType == ServerFormat::with_port) {
 		return server;
 	}
 
 	auto user = GetUser();
+	if (m_protocol == STORJ) {
+		// FIXME: The API key is not a user name. Move it into credentials section in a future version.
+		// Hide it here, we shouldn't display the key like this.
+		user.clear();
+	}
 	if (credentials.logonType_ != LogonType::anonymous) {
 		// For now, only escape if formatting for URL.
 		// Open question: Do we need some form of escapement for presentation within the GUI,
@@ -417,18 +435,6 @@ bool CServer::SetEncodingType(CharsetEncoding type, std::wstring const& encoding
 	return true;
 }
 
-bool CServer::SetCustomEncoding(std::wstring const& encoding)
-{
-	if (encoding.empty()) {
-		return false;
-	}
-
-	m_encodingType = ENCODING_CUSTOM;
-	m_customEncoding = encoding;
-
-	return true;
-}
-
 CharsetEncoding CServer::GetEncodingType() const
 {
 	return m_encodingType;
@@ -525,7 +531,7 @@ ServerProtocol CServer::GetProtocolFromPrefix(std::wstring const& prefix, Server
 	}
 
 	for (unsigned int i = 0; protocolInfos[i].protocol != UNKNOWN; ++i) {
-		if (protocolInfos[i].prefix == lower) {
+		if (protocolInfos[i].prefix == lower && protocolInfos[i].parse_from_prefix) {
 			return protocolInfos[i].protocol;
 		}
 	}
@@ -573,12 +579,21 @@ bool CServer::ProtocolHasFeature(ServerProtocol const protocol, ProtocolFeature
 		break;
 	case ProtocolFeature::Charset:
 		if (protocol == FTP || protocol == FTPS || protocol == FTPES || protocol == INSECURE_FTP ||
-			protocol == STORJ || protocol == SFTP) {
+			protocol == SFTP) {
 			return true;
 		}
 		break;
 	case ProtocolFeature::PreserveTimestamp:
+		if (protocol == FTP || protocol == FTPS || protocol == FTPES || protocol == INSECURE_FTP ||
+			protocol == SFTP ||
+			protocol == AZURE_BLOB || protocol == AZURE_FILE || protocol == B2 || protocol == BOX ||
+			protocol == DROPBOX || protocol == GOOGLE_CLOUD || protocol == GOOGLE_DRIVE ||
+			protocol == ONEDRIVE || protocol == S3 || protocol == SWIFT || protocol == WEBDAV) {
+			return true;
+		}
+		break;
 	case ProtocolFeature::ServerType:
+	case ProtocolFeature::UnixChmod:
 		if (protocol == FTP || protocol == FTPS || protocol == FTPES || protocol == INSECURE_FTP ||
 			protocol == SFTP) {
 			return true;
@@ -589,11 +604,6 @@ bool CServer::ProtocolHasFeature(ServerProtocol const protocol, ProtocolFeature
 			return true;
 		}
 		break;
-	case ProtocolFeature::S3Lifecycle:
-		if (protocol == S3) {
-			return true;
-		}
-		break;
 	case ProtocolFeature::RecursiveDelete:
 		if (protocol == GOOGLE_DRIVE || protocol == DROPBOX || protocol == ONEDRIVE || protocol == B2) {
 			return true;
@@ -606,17 +616,45 @@ bool CServer::ProtocolHasFeature(ServerProtocol const protocol, ProtocolFeature
 		}
 		break;
 	case ProtocolFeature::TemporaryUrl:
-		if (protocol == S3 || protocol == DROPBOX || protocol == AZURE_BLOB || protocol == AZURE_FILE) {
-			return true;
-		}
-		break;
-	case ProtocolFeature::S3Sse:
-		if (protocol == S3) {
+		if (protocol == B2 || protocol == S3 || protocol == DROPBOX || protocol == AZURE_BLOB || protocol == AZURE_FILE) {
 			return true;
 		}
 		break;
 	case ProtocolFeature::Security:
 		return protocol != HTTP && protocol != INSECURE_FTP && protocol != INSECURE_WEBDAV;
+	case ProtocolFeature::ProExclusive:
+		switch (protocol) {
+			case FTP:
+			case FTPS:
+			case FTPES:
+			case INSECURE_FTP:
+			case SFTP:
+			case HTTP:
+			case HTTPS:
+			case STORJ:
+			case STORJ_GRANT:
+				return false;
+			default:
+				return true;
+		}
+		break;
+	case ProtocolFeature::ListVersions:
+		if (protocol == B2 || protocol == BOX || protocol == DROPBOX ||
+			protocol == ONEDRIVE || protocol == GOOGLE_DRIVE || protocol == S3) {
+			return true;
+		}
+		break;
+	case ProtocolFeature::DownloadVersion:
+		if (protocol == B2 || protocol == BOX || protocol == DROPBOX ||
+			protocol == GOOGLE_DRIVE || protocol == S3) {
+			return true;
+		}
+		break;
+	case ProtocolFeature::DeleteVersion:
+		if (protocol == B2 || protocol == BOX || protocol == GOOGLE_DRIVE || protocol == S3) {
+			return true;
+		}
+		break;
 	}
 	return false;
 }
@@ -644,7 +682,7 @@ void CServer::ClearExtraParameters()
 	extraParameters_.clear();
 }
 
-std::wstring CServer::GetExtraParameter(std::string const& name) const
+std::wstring CServer::GetExtraParameter(std::string_view const& name) const
 {
 	auto it = extraParameters_.find(name);
 	if (it != extraParameters_.cend()) {
@@ -653,19 +691,27 @@ std::wstring CServer::GetExtraParameter(std::string const& name) const
 	return std::wstring();
 }
 
-std::map<std::string, std::wstring> const& CServer::GetExtraParameters() const
+std::map<std::string, std::wstring, std::less<>> const& CServer::GetExtraParameters() const
 {
 	return extraParameters_;
 }
 
-void CServer::SetExtraParameter(std::string const& name, std::wstring const& value)
+bool CServer::HasExtraParameter(std::string_view const& name) const
 {
+	return extraParameters_.find(name) != extraParameters_.cend();
+}
+
+void CServer::SetExtraParameter(std::string_view const& name, std::wstring const& value)
+{
+	auto it = extraParameters_.find(name);
 	if (value.empty()) {
-		extraParameters_.erase(name);
+		if (it != extraParameters_.cend()) {
+			extraParameters_.erase(it);
+		}
 	}
 	else {
 		bool found = false;
-		auto const& traits  = ExtraServerParameterTraits(m_protocol);
+		auto const& traits = ExtraServerParameterTraits(m_protocol);
 		for (auto const& trait : traits) {
 			if (trait.section_ != ParameterSection::credentials && name == trait.name_) {
 				found = true;
@@ -674,14 +720,22 @@ void CServer::SetExtraParameter(std::string const& name, std::wstring const& val
 		}
 
 		if (found) {
-			extraParameters_[name] = value;
+			if (it == extraParameters_.cend()) {
+				extraParameters_.emplace(name, value);
+			}
+			else {
+				it->second = value;
+			}
 		}
 	}
 }
 
-void CServer::ClearExtraParameter(const std::string &name)
+void CServer::ClearExtraParameter(std::string_view const& name)
 {
-	extraParameters_.erase(name);
+	auto it = extraParameters_.find(name);
+	if (it != extraParameters_.cend()) {
+		extraParameters_.erase(it);
+	}
 }
 
 LogonType GetLogonTypeFromName(std::wstring const& name)
@@ -701,6 +755,9 @@ LogonType GetLogonTypeFromName(std::wstring const& name)
 	else if (name == _("Account")) {
 		return LogonType::account;
 	}
+	else if (name == _("Profile")) {
+		return LogonType::profile;
+	}
 	else {
 		return LogonType::anonymous;
 	}
@@ -722,6 +779,8 @@ std::wstring GetNameFromLogonType(LogonType type)
 		return _("Interactive");
 	case LogonType::account:
 		return _("Account");
+	case LogonType::profile:
+		return _("Profile");
 	default:
 		return _("Anonymous");
 	}
@@ -749,7 +808,7 @@ void Credentials::ClearExtraParameters()
 	extraParameters_.clear();
 }
 
-std::wstring Credentials::GetExtraParameter(std::string const& name) const
+std::wstring Credentials::GetExtraParameter(std::string_view const& name) const
 {
 	auto it = extraParameters_.find(name);
 	if (it != extraParameters_.cend()) {
@@ -758,27 +817,51 @@ std::wstring Credentials::GetExtraParameter(std::string const& name) const
 	return std::wstring();
 }
 
-std::map<std::string, std::wstring> const& Credentials::GetExtraParameters() const
+std::map<std::string, std::wstring, std::less<>> const& Credentials::GetExtraParameters() const
 {
 	return extraParameters_;
 }
 
-void Credentials::SetExtraParameter(ServerProtocol protocol, std::string const& name, std::wstring const& value)
+bool Credentials::HasExtraParameter(std::string_view const& name) const
 {
-	bool found = false;
-	auto const& traits = ExtraServerParameterTraits(protocol);
-	for (auto const& trait : traits) {
-		if (trait.section_ != ParameterSection::credentials && name == trait.name_) {
-			found = true;
-			break;
+	return extraParameters_.find(name) != extraParameters_.cend();
+}
+
+void Credentials::SetExtraParameter(ServerProtocol protocol, std::string_view const& name, std::wstring const& value)
+{
+	auto it = extraParameters_.find(name);
+	if (value.empty()) {
+		if (it != extraParameters_.cend()) {
+			extraParameters_.erase(it);
 		}
 	}
+	else {
+		bool found = false;
+		auto const& traits = ExtraServerParameterTraits(protocol);
+		for (auto const& trait : traits) {
+			if (trait.section_ == ParameterSection::credentials && name == trait.name_) {
+				found = true;
+				break;
+			}
+		}
 
-	if (found) {
-		extraParameters_[name] = value;
+		if (found) {
+			if (it == extraParameters_.cend()) {
+				extraParameters_.emplace(name, value);
+			}
+			else {
+				it->second = value;
+			}
+		}
 	}
 }
 
+void Credentials::SetExtraParameters(ServerProtocol protocol, std::map<std::string, std::wstring, std::less<>> const& extraParameters)
+{
+	for (auto const& param : extraParameters) {
+		SetExtraParameter(protocol, param.first, param.second);
+	}
+}
 
 std::vector<LogonType> GetSupportedLogonTypes(ServerProtocol protocol)
 {
@@ -792,12 +875,14 @@ std::vector<LogonType> GetSupportedLogonTypes(ServerProtocol protocol)
 	case SFTP:
 		return {LogonType::anonymous, LogonType::normal, LogonType::ask, LogonType::interactive, LogonType::key};
 	case S3:
-		return {LogonType::anonymous, LogonType::normal, LogonType::ask};
+		return {LogonType::anonymous, LogonType::normal, LogonType::ask, LogonType::profile};
 	case STORJ:
+	case STORJ_GRANT:
 	case AZURE_FILE:
 	case AZURE_BLOB:
 	case SWIFT:
 	case B2:
+	case RACKSPACE:
 		return {LogonType::normal, LogonType::ask};
 	case WEBDAV:
 	case INSECURE_WEBDAV:
@@ -816,39 +901,125 @@ std::vector<LogonType> GetSupportedLogonTypes(ServerProtocol protocol)
 	return {LogonType::anonymous};
 }
 
+bool FZC_PUBLIC_SYMBOL IsSupportedLogonType(ServerProtocol protocol, LogonType type)
+{
+	auto const types = GetSupportedLogonTypes(protocol);
+	for (auto const& t : types) {
+		if (t == type) {
+			return true;
+		}
+	}
+	return false;
+}
+
 std::vector<ParameterTraits> const& ExtraServerParameterTraits(ServerProtocol protocol)
 {
 	switch (protocol) {
+	case FTP:
+	case FTPS:
+		{
+			static std::vector<ParameterTraits> const ret = []() {
+				std::vector<ParameterTraits> ret;
+				ret.emplace_back(ParameterTraits{"otp_code", ParameterSection::credentials, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+				return ret;
+			}();
+			return ret;
+		}
 	case GOOGLE_CLOUD:
 		{
-			static std::vector<ParameterTraits> ret = []() {
+			static std::vector<ParameterTraits> const ret = []() {
 				std::vector<ParameterTraits> ret;
-				ret.emplace_back(ParameterTraits{"email", ParameterSection::user, ParameterTraits::optional, std::wstring(), std::wstring()});
+				ret.emplace_back(ParameterTraits{"login_hint", ParameterSection::user, ParameterTraits::optional, std::wstring(), _("Name or email address")});
+				ret.emplace_back(ParameterTraits{"oauth_identity", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
 				return ret;
 			}();
 			return ret;
 		}
 	case SWIFT:
 		{
-			static std::vector<ParameterTraits> ret = []() {
+			static std::vector<ParameterTraits> const ret = []() {
 				std::vector<ParameterTraits> ret;
 				ret.emplace_back(ParameterTraits{"identpath", ParameterSection::host, 0, std::wstring(), _("Path of identity service")});
 				ret.emplace_back(ParameterTraits{"identuser", ParameterSection::user, ParameterTraits::optional, std::wstring(), std::wstring()});
+				ret.emplace_back(ParameterTraits{"keystone_version", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+				ret.emplace_back(ParameterTraits{"domain", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, L"Default", std::wstring()});
+				return ret;
+			}();
+			return ret;
+		}
+	case RACKSPACE:
+		{
+			static std::vector<ParameterTraits> const ret = []() {
+				std::vector<ParameterTraits> ret;
+				ret.emplace_back(ParameterTraits{"identpath", ParameterSection::host, 0, L"/v2.0/tokens", _("Path of identity service")});
+				ret.emplace_back(ParameterTraits{"identuser", ParameterSection::user, ParameterTraits::optional, std::wstring(), std::wstring()});
 				return ret;
 			}();
 			return ret;
 		}
 	case S3:
 		{
-			static std::vector<ParameterTraits> ret = []() {
+			static std::vector<ParameterTraits> const ret = []() {
 				std::vector<ParameterTraits> ret;
-				ret.emplace_back(ParameterTraits{"ssealgorithm", ParameterSection::custom, ParameterTraits::optional, std::wstring(), std::wstring()});
-				ret.emplace_back(ParameterTraits{"ssekmskey", ParameterSection::custom, ParameterTraits::optional, std::wstring(), std::wstring()});
-				ret.emplace_back(ParameterTraits{"ssecustomerkey", ParameterSection::custom, ParameterTraits::optional, std::wstring(), std::wstring()});
+				ret.emplace_back(ParameterTraits{"ssealgorithm", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+				ret.emplace_back(ParameterTraits{"ssekmskey", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+				ret.emplace_back(ParameterTraits{"ssecustomerkey", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+				ret.emplace_back(ParameterTraits{"stsrolearn", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+				ret.emplace_back(ParameterTraits{"stsmfaserial", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+				ret.emplace_back(ParameterTraits{"region", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+				ret.emplace_back(ParameterTraits{"original_profile", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
 				return ret;
 			}();
 			return ret;
 		}
+	case GOOGLE_DRIVE:
+	case ONEDRIVE:
+	{
+		static std::vector<ParameterTraits> const ret = []() {
+			std::vector<ParameterTraits> ret;
+			ret.emplace_back(ParameterTraits{"login_hint", ParameterSection::user, ParameterTraits::optional, std::wstring(), _("Name or email address")});
+			ret.emplace_back(ParameterTraits{"oauth_identity", ParameterSection::extra, ParameterTraits::optional| ParameterTraits::custom, std::wstring(), std::wstring()});
+			return ret;
+		}();
+		return ret;
+	}
+	case DROPBOX:
+	{
+		static std::vector<ParameterTraits> const ret = []() {
+			std::vector<ParameterTraits> ret;
+			ret.emplace_back(ParameterTraits{"oauth_identity", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+			ret.emplace_back(ParameterTraits{"root_namespace", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+			return ret;
+		}();
+		return ret;
+	}
+	case BOX:
+	{
+		static std::vector<ParameterTraits> const ret = []() {
+			std::vector<ParameterTraits> ret;
+			ret.emplace_back(ParameterTraits{"oauth_identity", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+			return ret;
+		}();
+		return ret;
+	}
+	case STORJ:
+	{
+		static std::vector<ParameterTraits> const ret = []() {
+			std::vector<ParameterTraits> ret;
+			ret.emplace_back(ParameterTraits{"passphrase_hash", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+			return ret;
+		}();
+		return ret;
+	}
+	case STORJ_GRANT:
+	{
+		static std::vector<ParameterTraits> const ret = []() {
+			std::vector<ParameterTraits> ret;
+			ret.emplace_back(ParameterTraits{"credentials_hash", ParameterSection::extra, ParameterTraits::optional | ParameterTraits::custom, std::wstring(), std::wstring()});
+			return ret;
+		}();
+		return ret;
+	}
 	default:
 		break;
 	}
@@ -879,6 +1050,11 @@ std::tuple<std::wstring, std::wstring> GetDefaultHost(ServerProtocol protocol)
 		return std::tuple<std::wstring, std::wstring>{L"api.backblazeb2.com", L""};
 	case BOX:
 		return std::tuple<std::wstring, std::wstring>{L"api.box.com", L""};
+	case RACKSPACE:
+		return std::tuple<std::wstring, std::wstring>{L"identity.api.rackspacecloud.com", L""};
+	case STORJ:
+	case STORJ_GRANT:
+		return std::tuple<std::wstring, std::wstring>{L"us-central-1.tardigrade.io", L""};
 	default:
 		break;
 	}
@@ -888,7 +1064,7 @@ std::tuple<std::wstring, std::wstring> GetDefaultHost(ServerProtocol protocol)
 
 bool ProtocolHasUser(ServerProtocol protocol)
 {
-	return protocol != DROPBOX && protocol != ONEDRIVE && protocol != BOX;
+	return protocol != DROPBOX && protocol != ONEDRIVE && protocol != BOX && protocol != GOOGLE_DRIVE && protocol != STORJ_GRANT;
 }
 
 bool CServer::SameResource(CServer const& other) const
@@ -896,10 +1072,23 @@ bool CServer::SameResource(CServer const& other) const
 	// We include post-login commands as it may be used for things like the HOST command.
 	// We include proxy parameters as a hostname may resolve to different servers depending on whether a proxy is used.
 
-	auto l = std::tie(m_protocol, m_host, m_port, m_user, m_postLoginCommands, m_bypassProxy, extraParameters_);
-	auto r = std::tie(other.m_protocol, other.m_host, other.m_port, other.m_user, other.m_postLoginCommands, other.m_bypassProxy, other.extraParameters_);
+	auto l = std::tie(m_protocol, m_host, m_port, m_user, m_postLoginCommands);
+	auto r = std::tie(other.m_protocol, other.m_host, other.m_port, other.m_user, other.m_postLoginCommands);
+	if (l != r) {
+		return false;
+	}
 
-	return l == r;
+	auto const& traits = ExtraServerParameterTraits(m_protocol);
+	for (auto const& trait : traits) {
+		if (trait.flags_ & ParameterTraits::content_transparent) {
+			continue;
+		}
+		if (GetExtraParameter(trait.name_) != other.GetExtraParameter(trait.name_)) {
+			return false;
+		}
+	}
+
+	return true;
 }
 
 bool CServer::SameContent(CServer const& other) const
