@@ -1,10 +1,10 @@
-#include <filezilla.h>
+#include "filezilla.h"
 
 #include "dndobjects.h"
 #include "dragdropmanager.h"
 #include "drop_target_ex.h"
 #include "filezillaapp.h"
-#include "filter.h"
+#include "filter_manager.h"
 #include "file_utils.h"
 #include "graphics.h"
 #include "inputdialog.h"
@@ -15,8 +15,9 @@
 
 #include <libfilezilla/local_filesys.hpp>
 
+#include <wx/menu.h>
+
 #ifdef __WXMSW__
-#include <wx/msw/registry.h>
 #include <shlobj.h>
 #include <dbt.h>
 #include "volume_enumerator.h"
@@ -28,6 +29,8 @@
 
 #include <algorithm>
 
+using namespace std::literals;
+
 class CTreeItemData : public wxTreeItemData
 {
 public:
@@ -44,15 +47,6 @@ public:
 	{
 	}
 
-	void ClearDropHighlight()
-	{
-		const wxTreeItemId dropHighlight = m_pLocalTreeView->m_dropHighlight;
-		if (dropHighlight != wxTreeItemId()) {
-			m_pLocalTreeView->SetItemDropHighlight(dropHighlight, false);
-			m_pLocalTreeView->m_dropHighlight = wxTreeItemId();
-		}
-	}
-
 	std::wstring GetDirFromItem(wxTreeItemId const& item)
 	{
 		std::wstring const dir = m_pLocalTreeView->GetDirFromItem(item);
@@ -68,7 +62,7 @@ public:
 
 	wxTreeItemId GetHit(const wxPoint& point)
 	{
-		int flags = 0;
+		int flags{};
 		wxTreeItemId hit = m_pLocalTreeView->HitTest(point, flags);
 
 		if (flags & (wxTREE_HITTEST_ABOVE | wxTREE_HITTEST_BELOW | wxTREE_HITTEST_NOWHERE | wxTREE_HITTEST_TOLEFT | wxTREE_HITTEST_TORIGHT)) {
@@ -109,26 +103,27 @@ public:
 			pDragDropManager->pDropTarget = m_pLocalTreeView;
 		}
 
-		auto const format = m_pDataObject->GetReceivedFormat();
+		auto const format = GetReceivedFormat();
 		if (format == m_pFileDataObject->GetFormat()) {
 			m_pLocalTreeView->m_state.HandleDroppedFiles(m_pFileDataObject, path, def == wxDragCopy);
 		}
-		else if (format == m_pLocalDataObject->GetFormat()) {
-			m_pLocalTreeView->m_state.HandleDroppedFiles(m_pLocalDataObject, path, def == wxDragCopy);
+		else if (format == LocalDataObjectFormat()) {
+			m_pLocalTreeView->m_state.HandleDroppedFiles(GetLocalDataObject(), path, def == wxDragCopy);
 		}
 		else {
-			if (m_pRemoteDataObject->GetProcessId() != (int)wxGetProcessId()) {
+			auto * obj = GetRemoteDataObject();
+			if (obj->GetProcessId() != (int)wxGetProcessId()) {
 				wxMessageBoxEx(_("Drag&drop between different instances of FileZilla has not been implemented yet."));
 				return wxDragNone;
 			}
 
 			auto & state = m_pLocalTreeView->m_state;
-			if (!state.GetSite() || m_pRemoteDataObject->GetSite().server != state.GetSite().server) {
+			if (!state.GetSite() || obj->GetSite().server != state.GetSite().server) {
 				wxMessageBoxEx(_("Drag&drop between different servers has not been implemented yet."));
 				return wxDragNone;
 			}
 
-			if (!state.DownloadDroppedFiles(m_pRemoteDataObject, path)) {
+			if (!state.DownloadDroppedFiles(obj, path)) {
 				return wxDragNone;
 			}
 		}
@@ -138,11 +133,11 @@ public:
 
 	virtual bool OnDrop(wxCoord x, wxCoord y)
 	{
-		if (!CScrollableDropTarget<wxTreeCtrlEx>::OnDrop(x, y)) {
+		if (!CFileDropTarget::OnDrop(x, y)) {
 			return false;
 		}
 
-		ClearDropHighlight();
+		m_pLocalTreeView->ClearDropHighlight();
 
 		wxTreeItemId hit = GetHit(wxPoint(x, y));
 		if (!hit) {
@@ -157,41 +152,36 @@ public:
 		return true;
 	}
 
-	wxTreeItemId DisplayDropHighlight(wxPoint point)
+	wxTreeItemId DisplayDropHighlight(wxPoint const& point) override
 	{
 		wxTreeItemId hit = GetHit(point);
 		if (!hit) {
-			ClearDropHighlight();
+			m_pLocalTreeView->ClearDropHighlight();
 			return hit;
 		}
 
 		std::wstring dir = GetDirFromItem(hit);
 
 		if (dir.empty()) {
-			ClearDropHighlight();
+			m_pLocalTreeView->ClearDropHighlight();
 			return wxTreeItemId();
 		}
-
-		const wxTreeItemId dropHighlight = m_pLocalTreeView->m_dropHighlight;
-		if (dropHighlight != wxTreeItemId()) {
-			m_pLocalTreeView->SetItemDropHighlight(dropHighlight, false);
+		else {
+			m_pLocalTreeView->DisplayDropHighlight(hit);
 		}
 
-		m_pLocalTreeView->SetItemDropHighlight(hit, true);
-		m_pLocalTreeView->m_dropHighlight = hit;
-
 		return hit;
 	}
 
 	virtual wxDragResult OnDragOver(wxCoord x, wxCoord y, wxDragResult def)
 	{
-		def = CScrollableDropTarget<wxTreeCtrlEx>::OnDragOver(x, y, def);
+		def = CScrollableDropTarget::OnDragOver(x, y, def);
 
 		if (def == wxDragError ||
 			def == wxDragNone ||
 			def == wxDragCancel)
 		{
-			ClearDropHighlight();
+			m_pLocalTreeView->ClearDropHighlight();
 			return def;
 		}
 
@@ -209,13 +199,13 @@ public:
 
 	virtual void OnLeave()
 	{
-		CScrollableDropTarget<wxTreeCtrlEx>::OnLeave();
-		ClearDropHighlight();
+		CScrollableDropTarget::OnLeave();
+		m_pLocalTreeView->ClearDropHighlight();
 	}
 
 	virtual wxDragResult OnEnter(wxCoord x, wxCoord y, wxDragResult def)
 	{
-		def = CScrollableDropTarget<wxTreeCtrlEx>::OnEnter(x, y, def);
+		def = CScrollableDropTarget::OnEnter(x, y, def);
 		return OnDragOver(x, y, def);
 	}
 
@@ -247,13 +237,15 @@ EVT_CHAR(CLocalTreeView::OnChar)
 EVT_MENU(XRCID("ID_OPEN"), CLocalTreeView::OnMenuOpen)
 END_EVENT_TABLE()
 
-CLocalTreeView::CLocalTreeView(wxWindow* parent, wxWindowID id, CState& state, CQueueView *pQueueView)
-	: wxTreeCtrlEx(parent, id, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL | wxTR_EDIT_LABELS | wxTR_LINES_AT_ROOT | wxTR_HAS_BUTTONS | wxNO_BORDER),
-	CSystemImageList(CThemeProvider::GetIconSize(iconSizeSmall).x),
-	CStateEventHandler(state),
-	m_pQueueView(pQueueView)
+CLocalTreeView::CLocalTreeView(wxWindow* parent, wxWindowID id, CState& state, CQueueView *pQueueView, COptionsBase & options)
+	: wxTreeCtrlEx(parent, id, wxDefaultPosition, wxDefaultSize, DEFAULT_TREE_STYLE | wxTAB_TRAVERSAL | wxTR_EDIT_LABELS | wxNO_BORDER)
+	, CSystemImageList(CThemeProvider::GetIconSize(iconSizeSmall).x)
+	, CStateEventHandler(state)
+	, COptionChangeEventHandler(this)
+	, m_pQueueView(pQueueView)
+    , options_(options)
 {
-	wxGetApp().AddStartupProfileRecord("CLocalTreeView::CLocalTreeView");
+	wxGetApp().AddStartupProfileRecord("CLocalTreeView::CLocalTreeView"sv);
 #ifdef __WXMAC__
 	SetFont(wxSystemSettings::GetFont(wxSYS_DEFAULT_GUI_FONT));
 #endif
@@ -265,7 +257,7 @@ CLocalTreeView::CLocalTreeView(wxWindow* parent, wxWindowID id, CState& state, C
 	SetImageList(GetSystemImageList());
 
 	UpdateSortMode();
-	RegisterOption(OPTION_FILELIST_NAMESORT);
+	options_.watch(OPTION_FILELIST_NAMESORT, this);
 
 #ifdef __WXMSW__
 	m_pVolumeEnumeratorThread = 0;
@@ -295,6 +287,7 @@ CLocalTreeView::CLocalTreeView(wxWindow* parent, wxWindowID id, CState& state, C
 
 CLocalTreeView::~CLocalTreeView()
 {
+	options_.unwatch_all(this);
 #ifdef __WXMSW__
 	delete m_pVolumeEnumeratorThread;
 #endif
@@ -424,7 +417,7 @@ wxTreeItemId CLocalTreeView::GetSubdir(wxTreeItemId parent, wxString const& subD
 
 bool CLocalTreeView::DisplayDrives(wxTreeItemId parent)
 {
-	wxGetApp().AddStartupProfileRecord("CLocalTreeView::DisplayDrives");
+	wxGetApp().AddStartupProfileRecord("CLocalTreeView::DisplayDrives"sv);
 
 	std::vector<std::wstring> drives = CVolumeDescriptionEnumeratorThread::GetDrives();
 
@@ -446,7 +439,7 @@ bool CLocalTreeView::DisplayDrives(wxTreeItemId parent)
 	}
 	SortChildren(parent);
 
-	wxGetApp().AddStartupProfileRecord("CLocalTreeView::DisplayDrives adding drives done");
+	wxGetApp().AddStartupProfileRecord("CLocalTreeView::DisplayDrives adding drives done"sv);
 
 	return true;
 }
@@ -492,15 +485,14 @@ void CLocalTreeView::DisplayDir(wxTreeItemId parent, std::wstring const& dirname
 	bool matchedKnown = false;
 
 	fz::native_string file;
-	bool wasLink;
-	int attributes;
-	bool is_dir;
+	bool wasLink{};
+	int attributes{};
+	fz::local_filesys::type t{};
 	static int64_t const size(-1);
 	fz::datetime date;
 
 	wxTreeItemId last;
-	while (local_filesys.get_next_file(file, wasLink, is_dir, 0, &date, &attributes)) {
-		wxASSERT(is_dir);
+	while (local_filesys.get_next_file(file, wasLink, t, 0, &date, &attributes)) {
 		std::wstring wfile = fz::to_wstring(file);
 		if (file.empty() || wfile.empty()) {
 			wxGetApp().DisplayEncodingWarning();
@@ -561,13 +553,13 @@ std::wstring CLocalTreeView::HasSubdir(std::wstring const& dirname)
 	}
 
 	fz::native_string file;
-	bool wasLink;
-	int attributes;
-	bool is_dir;
+	bool wasLink{};
+	int attributes{};
+	fz::local_filesys::type t{};
 	static int64_t const size(-1);
 	fz::datetime date;
-	while (local_filesys.get_next_file(file, wasLink, is_dir, 0, &date, &attributes)) {
-		wxASSERT(is_dir);
+	while (local_filesys.get_next_file(file, wasLink, t, 0, &date, &attributes)) {
+
 		std::wstring wfile = fz::to_wstring(file);
 		if (file.empty() || wfile.empty()) {
 			wxGetApp().DisplayEncodingWarning();
@@ -638,26 +630,28 @@ std::wstring CLocalTreeView::GetDirFromItem(wxTreeItemId item)
 	while (item) {
 #ifdef __WXMSW__
 		if (item == m_desktop) {
-			wxChar path[MAX_PATH + 1];
-			if (SHGetFolderPath(0, CSIDL_DESKTOPDIRECTORY, 0, SHGFP_TYPE_CURRENT, path) != S_OK) {
-				if (SHGetFolderPath(0, CSIDL_DESKTOP, 0, SHGFP_TYPE_CURRENT, path) != S_OK) {
-					wxMessageBoxEx(_("Failed to get desktop path"));
-					return L"/";
-				}
+			wchar_t* out{};
+			if (SHGetKnownFolderPath(FOLDERID_Desktop, 0, 0, &out) != S_OK) {
+				wxMessageBoxEx(_("Failed to get desktop path"));
+				return L"/";
 			}
-			dir = path;
+			dir = out;
+			CoTaskMemFree(out);
+
 			if (dir.empty() || dir.back() != separator) {
 				dir += separator;
 			}
 			return dir;
 		}
 		else if (item == m_documents) {
-			wxChar path[MAX_PATH + 1];
-			if (SHGetFolderPath(0, CSIDL_PERSONAL, 0, SHGFP_TYPE_CURRENT, path) != S_OK) {
+			wchar_t* out{};
+			if (SHGetKnownFolderPath(FOLDERID_Documents, 0, 0, &out) != S_OK) {
 				wxMessageBoxEx(_("Failed to get 'My Documents' path"));
 				return L"/";
 			}
-			dir = path;
+			dir = out;
+			CoTaskMemFree(out);
+
 			if (dir.empty() || dir.back() != separator) {
 				dir += separator;
 			}
@@ -694,18 +688,18 @@ std::wstring CLocalTreeView::GetDirFromItem(wxTreeItemId item)
 
 void CLocalTreeView::UpdateSortMode()
 {
-	CFileListCtrlSortBase::NameSortMode sortMode;
-	switch (COptions::Get()->GetOptionVal(OPTION_FILELIST_NAMESORT))
+	NameSortMode sortMode;
+	switch (options_.get_int(OPTION_FILELIST_NAMESORT))
 	{
 	case 0:
 	default:
-		sortMode = CFileListCtrlSortBase::namesort_caseinsensitive;
+		sortMode = NameSortMode::case_insensitive;
 		break;
 	case 1:
-		sortMode = CFileListCtrlSortBase::namesort_casesensitive;
+		sortMode = NameSortMode::case_sensitive;
 		break;
 	case 2:
-		sortMode = CFileListCtrlSortBase::namesort_natural;
+		sortMode = NameSortMode::natural;
 		break;
 	}
 	sortFunction_ = CFileListCtrlSortBase::GetCmpFunction(sortMode);
@@ -737,7 +731,7 @@ void CLocalTreeView::RefreshListing()
 	for (auto child = GetFirstChild(m_drives, tmp); child; child = GetNextSibling(child)) {
 		if (IsExpanded(child)) {
 			std::wstring drive = GetItemText(child).ToStdWstring();
-			size_t pos = drive.find(' ' );
+			size_t pos = drive.find(' ');
 			if (pos != std::wstring::npos) {
 				drive = drive.substr(0, pos);
 			}
@@ -784,11 +778,11 @@ void CLocalTreeView::RefreshListing()
 
 		fz::native_string file;
 		static int64_t const size(-1);
-		bool was_link;
-		bool is_dir;
-		int attributes;
+		bool was_link{};
+		fz::local_filesys::type t{};
+		int attributes{};
 		fz::datetime date;
-		while (local_filesys.get_next_file(file, was_link, is_dir, 0, &date, &attributes)) {
+		while (local_filesys.get_next_file(file, was_link, t, 0, &date, &attributes)) {
 			std::wstring wfile = fz::to_wstring(file);
 			if (file.empty() || wfile.empty()) {
 				wxGetApp().DisplayEncodingWarning();
@@ -806,19 +800,17 @@ void CLocalTreeView::RefreshListing()
 		// Step 3: Merge list of subdirectories with subtree.
 		std::vector<wxTreeItemId> toDelete;
 
-		bool inserted = false;
-
 		wxTreeItemIdValue unused;
 		wxTreeItemId child = GetFirstChild(dir.item, unused);
 
+		wxTreeItemId firstInserted{};
 		wxTreeItemId last = GetLastChild(dir.item);
-
 		auto iter = dirs.begin();
 		while (child || iter != dirs.end()) {
 			int cmp;
 			if (child && iter != dirs.end()) {
 				wxString const& childName = GetItemText(child);
-				cmp = sortFunction_(std::wstring_view(childName.data(), childName.size()), *iter);
+				cmp = sortFunction_(*iter, std::wstring_view(childName.data(), childName.size()));
 			}
 			else if (child) {
 				cmp = 1;
@@ -845,6 +837,9 @@ void CLocalTreeView::RefreshListing()
 					dirsToCheck.push_front(subdir);
 				}
 				child = GetNextSibling(child);
+				if (child == firstInserted) {
+					child = wxTreeItemId{};
+				}
 				++iter;
 			}
 			else if (cmp > 0) {
@@ -855,8 +850,13 @@ void CLocalTreeView::RefreshListing()
 				while (sel && sel != child) {
 					sel = GetItemParent(sel);
 				}
-				toDelete.push_back(child);
+				if (!sel) {
+					toDelete.push_back(child);
+				}
 				child = GetNextSibling(child);
+				if (child == firstInserted) {
+					child = wxTreeItemId{};
+				}
 			}
 			else if (cmp < 0) {
 				// New subdirectory, add treeitem
@@ -869,15 +869,18 @@ void CLocalTreeView::RefreshListing()
 #endif
 					);
 
+				if (!firstInserted) {
+					firstInserted = last;
+				}
+
 				CheckSubdirStatus(last, fullname);
 				++iter;
-				inserted = true;
 			}
 		}
 		for (auto it = toDelete.rbegin(); it != toDelete.rend(); ++it) {
 			Delete(*it);
 		}
-		if (inserted) {
+		if (firstInserted) {
 			SortChildren(dir.item);
 		}
 	}
@@ -931,7 +934,7 @@ void CLocalTreeView::OnStateChange(t_statechange_notifications notification, std
 
 void CLocalTreeView::OnBeginDrag(wxTreeEvent& event)
 {
-	if (COptions::Get()->GetOptionVal(OPTION_DND_DISABLED) != 0) {
+	if (options_.get_int(OPTION_DND_DISABLED) != 0) {
 		return;
 	}
 
@@ -966,21 +969,22 @@ void CLocalTreeView::OnBeginDrag(wxTreeEvent& event)
 	}
 #endif
 
-	CDragDropManager* pDragDropManager = CDragDropManager::Init();
-	pDragDropManager->pDragSource = this;
-
 #ifdef __WXMAC__
 	// Don't use wxFileDataObject on Mac, crashes on Mojave, wx bug #18232
 	CLocalDataObject obj;
-	obj.AddFile(dir);
 #else
 	wxFileDataObject obj;
-	obj.AddFile(dir);
 #endif
 
-	wxDropSource source(this);
+	CDragDropManager* pDragDropManager = CDragDropManager::Init();
+	pDragDropManager->pDragSource = this;
+
+	obj.AddFile(dir);
+	pDragDropManager->dragDataObject = &obj;
+
+	DropSource source(this);
 	source.SetData(obj);
-	int res = source.DoDragDrop(wxDrag_AllowMove);
+	int res = source.DoFileDragDrop(wxDrag_AllowMove);
 
 	bool handled_internally = pDragDropManager->pDropTarget != 0;
 
@@ -991,6 +995,13 @@ void CLocalTreeView::OnBeginDrag(wxTreeEvent& event)
 		// externally, the internal handlers do this for us already
 		m_state.RefreshLocal();
 	}
+
+#ifdef __WXMAC__
+	if (!source.m_OutDir.IsEmpty()) {
+		CLocalPath target(source.m_OutDir.ToStdWstring());
+		m_state.HandleDroppedFiles(&obj, target, true);
+	}
+#endif
 }
 
 #ifdef __WXMSW__
@@ -1104,10 +1115,10 @@ void CLocalTreeView::OnContextMenu(wxTreeEvent& event)
 
 	wxMenu menu;
 	auto item = new wxMenuItem(&menu, XRCID("ID_UPLOAD"), _("&Upload"), _("Upload selected directory"));
-	item->SetBitmap(wxArtProvider::GetBitmap(_T("ART_UPLOAD"), wxART_MENU));
+	item->SetBitmap(MakeBmpBundle(wxArtProvider::GetBitmap(_T("ART_UPLOAD"), wxART_MENU)));
 	menu.Append(item);
 	item = new wxMenuItem(&menu, XRCID("ID_ADDTOQUEUE"), _("&Add to queue"), _("Add selected directory to the transfer queue"));
-	item->SetBitmap(wxArtProvider::GetBitmap(_T("ART_UPLOADADD"), wxART_MENU));
+	item->SetBitmap(MakeBmpBundle(wxArtProvider::GetBitmap(_T("ART_UPLOADADD"), wxART_MENU)));
 	menu.Append(item);
 
 	menu.AppendSeparator();
@@ -1181,13 +1192,13 @@ void CLocalTreeView::OnMenuUpload(wxCommandEvent& event)
 	bool const queue_only = event.GetId() == XRCID("ID_ADDTOQUEUE");
 
 	CFilterManager filter;
-	recursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_transfer, filter.GetActiveFilters(), !queue_only);
+	recursiveOperation->StartRecursiveOperation(recursive_operation::recursive_transfer, filter.GetActiveFilters(), !queue_only);
 }
 
 // Create a new Directory
 void CLocalTreeView::OnMenuMkdir(wxCommandEvent&)
 {
-	wxString newdir = MenuMkdir();
+	CLocalPath newdir = MenuMkdir();
 	if (!newdir.empty()) {
 		RefreshListing();
 		m_state.RefreshLocal();
@@ -1197,7 +1208,7 @@ void CLocalTreeView::OnMenuMkdir(wxCommandEvent&)
 // Create a new Directory and enter the new Directory
 void CLocalTreeView::OnMenuMkdirChgDir(wxCommandEvent&)
 {
-	std::wstring newdir = MenuMkdir().ToStdWstring();
+	CLocalPath newdir = MenuMkdir();
 	if (newdir.empty()) {
 		return;
 	}
@@ -1216,53 +1227,42 @@ void CLocalTreeView::OnMenuMkdirChgDir(wxCommandEvent&)
 
 // Helper-Function to create a new Directory
 // Returns the name of the new directory
-wxString CLocalTreeView::MenuMkdir()
+CLocalPath CLocalTreeView::MenuMkdir()
 {
 	if (!m_contextMenuItem.IsOk()) {
-		return wxString();
+		return {};
 	}
 
-	std::wstring path = GetDirFromItem(m_contextMenuItem);
-	if (!path.empty() && path.back() != wxFileName::GetPathSeparator()) {
-		path += wxFileName::GetPathSeparator();
-	}
-
-	if (!CLocalPath(path).IsWriteable()) {
+	CLocalPath p(GetDirFromItem(m_contextMenuItem));
+	if (p.empty() || !p.IsWriteable()) {
 		wxBell();
-		return wxString();
+		return {};
 	}
 
 	CInputDialog dlg;
 	if (!dlg.Create(this, _("Create directory"), _("Please enter the name of the directory which should be created:"))) {
-		return wxString();
+		return {};
 	}
 
 	wxString newName = _("New directory");
-	dlg.SetValue(path + newName);
-	dlg.SelectText(path.size(), path.size() + newName.size());
+	dlg.SetValue(p.GetPath() + newName);
+	dlg.SelectText(p.GetPath().size(), p.GetPath().size() + newName.size());
 
 	if (dlg.ShowModal() != wxID_OK) {
-		return wxString();
+		return {};
 	}
 
-	wxFileName fn(dlg.GetValue(), _T(""));
-	if (!fn.Normalize(wxPATH_NORM_ALL, path)) {
+	if (!p.ChangePath(dlg.GetValue().ToStdWstring()) || !p.IsWriteable()) {
 		wxBell();
-		return wxString();
+		return {};
 	}
 
-	bool res;
-	{
-		wxLogNull log;
-		res = fn.Mkdir(fn.GetPath(), 0777, wxPATH_MKDIR_FULL);
-	}
-
-	if (!res) {
+	if (!fz::mkdir(fz::to_native(p.GetPath()), true)) {
 		wxBell();
-		return wxString();
+		return {};
 	}
 
-	return fn.GetPath();
+	return p;
 }
 
 void CLocalTreeView::OnMenuRename(wxCommandEvent&)
@@ -1567,7 +1567,7 @@ wxTreeItemId CLocalTreeView::AddDrive(wxChar letter)
 	drive += _T(":");
 
 	long drivesToHide = CVolumeDescriptionEnumeratorThread::GetDrivesToHide();
-	if( CVolumeDescriptionEnumeratorThread::IsHidden(drive.wc_str(), drivesToHide) ) {
+	if ( CVolumeDescriptionEnumeratorThread::IsHidden(drive.wc_str(), drivesToHide) ) {
 		return wxTreeItemId();
 	}
 
@@ -1663,7 +1663,7 @@ void CLocalTreeView::OnMenuOpen(wxCommandEvent&)
 	OpenInFileManager(path);
 }
 
-void CLocalTreeView::OnOptionsChanged(changed_options_t const& options)
+void CLocalTreeView::OnOptionsChanged(watched_options const& options)
 {
 	if (options.test(OPTION_FILELIST_NAMESORT)) {
 		UpdateSortMode();
