@@ -276,13 +276,13 @@ void ShowOverlay(COptions& options, std::wstring const& data, wxTopLevelWindow*
 
 	auto box = new wxBoxSizer(wxVERTICAL);
 	p->SetSizer(box);
-	
+
 	auto outer = lay.createFlex(1);
 	box->Add(outer, 0, wxALL, 7);
 
 	auto sides = lay.createFlex(0, 1);
 	outer->Add(sides);
-	
+
 	auto right = lay.createFlex(1);
 	sides->Add(right, lay.valign);
 	if (tokens.size() > 4) {
@@ -315,7 +315,7 @@ void ShowOverlay(COptions& options, std::wstring const& data, wxTopLevelWindow*
 			}
 			if (bold) {
 				wxFont f = ctrl->GetFont();
-				f.SetWeight(wxBOLD);
+				f.SetWeight(wxFONTWEIGHT_BOLD);
 				ctrl->SetFont(f);
 			}
 			right->Add(ctrl);
@@ -502,7 +502,7 @@ CMainFrame::CMainFrame(COptions& options)
 #endif
 
 	wxGetApp().AddStartupProfileRecord("CMainFrame::CMainFrame pre layout"sv);
-	m_pWindowStateManager = new CWindowStateManager(this);
+	m_pWindowStateManager = new CWindowStateManager(this, options_);
 	m_pWindowStateManager->Restore(OPTION_MAINWINDOW_POSITION);
 
 	Layout();
@@ -2365,12 +2365,12 @@ void CMainFrame::SetDefaultSplitterPositions()
 		m_pTopSplitter->SetSashPosition(97);
 	}
 
-	wxSize size = m_pBottomSplitter->GetClientSize();
-	int h = size.GetHeight() - 135;
-	if (h < 50) {
-		h = 50;
-	}
 	if (m_pBottomSplitter) {
+		wxSize size = m_pBottomSplitter->GetClientSize();
+		int h = size.GetHeight() - 135;
+		if (h < 50) {
+			h = 50;
+		}
 		m_pBottomSplitter->SetSashPosition(h);
 	}
 
@@ -2406,16 +2406,6 @@ void CMainFrame::OnActivate(wxActivateEvent& event)
 		return;
 	}
 
-#if defined(__WXMAC__) && !wxCHECK_VERSION(3, 1, 0)
-	// wxMac looses focus information if the window becomes inactive.
-	// Restore focus to the previously focused child, otherwise focus ends up
-	// in the quickconnect bar.
-	// Go via ID of the last focused child to avoid issues with window lifetime.
-	if (m_lastFocusedChild != -1) {
-		m_winLastFocused = FindWindow(m_lastFocusedChild);
-	}
-#endif
-
 	CEditHandler* pEditHandler = CEditHandler::Get();
 	if (pEditHandler) {
 		pEditHandler->CheckForModifications();
@@ -2939,17 +2929,6 @@ void CMainFrame::SetupKeyboardAccelerators()
 #ifdef __WXMAC__
 	entries.emplace_back(wxACCEL_CMD, ',', XRCID("wxID_PREFERENCES"));
 
-#if !wxCHECK_VERSION(3, 2,1)
-	keyboardCommands[wxNewId()] = std::make_pair([](wxTextEntry* e) { e->Cut(); }, 'X');
-	keyboardCommands[wxNewId()] = std::make_pair([](wxTextEntry* e) { e->Copy(); }, 'C');
-	keyboardCommands[wxNewId()] = std::make_pair([](wxTextEntry* e) { e->Paste(); }, 'V');
-	keyboardCommands[wxNewId()] = std::make_pair([](wxTextEntry* e) { e->SelectAll(); }, 'A');
-
-	for (auto const& command : keyboardCommands) {
-		entries.emplace_back(wxACCEL_CMD, command.second.second, command.first);
-	}
-#endif
-
 	// Ctrl+(Shift+)Tab to switch between tabs
 	int id = wxNewId();
 	entries.emplace_back(wxACCEL_RAW_CTRL, '\t', id);
