@@ -1,14 +1,16 @@
-#include <filezilla.h>
+#include "filezilla.h"
 #include "bookmarks_dialog.h"
 #include "filezillaapp.h"
 #include "sitemanager.h"
-#include "ipcmutex.h"
 #include "state.h"
 #include "themeprovider.h"
 #include "treectrlex.h"
 #include "xmlfunctions.h"
 #include "xrc_helper.h"
 
+#include "../commonui/ipcmutex.h"
+
+#include <wx/dirdlg.h>
 #include <wx/statline.h>
 
 BEGIN_EVENT_TABLE(CNewBookmarkDialog, wxDialogEx)
@@ -29,14 +31,15 @@ EVT_TREE_BEGIN_LABEL_EDIT(XRCID("ID_TREE"), CBookmarksDialog::OnBeginLabelEdit)
 EVT_TREE_END_LABEL_EDIT(XRCID("ID_TREE"), CBookmarksDialog::OnEndLabelEdit)
 END_EVENT_TABLE()
 
-CNewBookmarkDialog::CNewBookmarkDialog(wxWindow* parent, std::wstring& site_path, Site const* site)
+CNewBookmarkDialog::CNewBookmarkDialog(wxWindow* parent, COptionsBase & options, std::wstring& site_path, Site const* site)
 	: m_parent(parent)
+	, options_(options)
 	, m_site_path(site_path)
 	, site_(site)
 {
 }
 
-int CNewBookmarkDialog::Run(const wxString &local_path, const CServerPath &remote_path)
+int CNewBookmarkDialog::Run(wxString const& local_path, CServerPath const& remote_path)
 {
 	if (!Load(m_parent, _T("ID_NEWBOOKMARK"))) {
 		return wxID_CANCEL;
@@ -70,7 +73,7 @@ void CNewBookmarkDialog::OnOK(wxCommandEvent&)
 	CServerPath remote_path;
 	if (!remote_path_raw.empty()) {
 		if (!global && site_) {
-			remote_path.SetType(site_->server.GetType());
+			remote_path.SetType(site_->GetOriginalServer().GetType());
 		}
 		if (!remote_path.SetPath(remote_path_raw.ToStdWstring())) {
 			wxMessageBoxEx(_("Could not parse remote path."), _("New bookmark"), wxICON_EXCLAMATION);
@@ -94,7 +97,7 @@ void CNewBookmarkDialog::OnOK(wxCommandEvent&)
 	if (!global && site_) {
 		std::unique_ptr<Site> site;
 		if (!m_site_path.empty()) {
-			site = CSiteManager::GetSiteByPath(m_site_path).first;
+			site = CSiteManager::GetSiteByPath(options_, m_site_path).first;
 		}
 		if (!site) {
 			if (wxMessageBoxEx(_("Site-specific bookmarks require the server to be stored in the Site Manager.\nAdd current connection to the site manager?"), _("New bookmark"), wxYES_NO | wxICON_QUESTION, this) != wxYES) {
@@ -159,8 +162,9 @@ public:
 	bool m_comparison{};
 };
 
-CBookmarksDialog::CBookmarksDialog(wxWindow* parent, std::wstring& site_path, Site const* site)
+CBookmarksDialog::CBookmarksDialog(wxWindow* parent, COptionsBase & options, std::wstring& site_path, Site const* site)
 	: m_parent(parent)
+	, options_(options)
 	, m_site_path(site_path)
 	, site_(site)
 {
@@ -223,7 +227,7 @@ void CBookmarksDialog::LoadSiteSpecificBookmarks()
 		return;
 	}
 
-	auto const site = CSiteManager::GetSiteByPath(m_site_path).first;
+	auto const site = CSiteManager::GetSiteByPath(options_, m_site_path).first;
 	if (!site) {
 		return;
 	}
@@ -261,7 +265,7 @@ int CBookmarksDialog::Run()
 
 	left->Add(new wxStaticText(this, wxID_ANY, _("Bookmark:")));
 
-	tree_ = new wxTreeCtrlEx(this, XRCID("ID_TREE"), wxDefaultPosition, wxDefaultSize, wxBORDER_SUNKEN | wxTR_EDIT_LABELS | wxTR_HAS_BUTTONS | wxTR_HIDE_ROOT);
+	tree_ = new wxTreeCtrlEx(this, XRCID("ID_TREE"), wxDefaultPosition, wxDefaultSize, DEFAULT_TREE_STYLE | wxBORDER_SUNKEN | wxTR_EDIT_LABELS | wxTR_HIDE_ROOT);
 	tree_->SetFocus();
 	tree_->SetMinSize(wxSize(-1, 250));
 	left->Add(tree_, lay.grow)->SetProportion(1);
@@ -369,7 +373,7 @@ void CBookmarksDialog::SaveGlobalBookmarks()
 		}
 	}
 
-	if (!file.Save(false)) {
+	if (!file.Save()) {
 		wxString msg = wxString::Format(_("Could not write \"%s\", the global bookmarks could no be saved: %s"), file.GetFileName(), file.GetError());
 		wxMessageBoxEx(msg, _("Error writing xml file"), wxICON_ERROR);
 	}
@@ -474,14 +478,14 @@ bool CBookmarksDialog::Verify()
 	if (!remotePathRaw.empty()) {
 		CServerPath remotePath;
 		if (site) {
-			remotePath.SetType(site->server.GetType());
+			remotePath.SetType(site->GetOriginalServer().GetType());
 		}
 		if (!remotePath.SetPath(remotePathRaw.ToStdWstring())) {
 			xrc_call(*this, "ID_BOOKMARK_REMOTEDIR", &wxTextCtrl::SetFocus);
 			if (site) {
 				wxString msg;
-				if (site->server.GetType() != DEFAULT) {
-					msg = wxString::Format(_("Remote path cannot be parsed. Make sure it is a valid absolute path and is supported by the current site's servertype (%s)."), CServer::GetNameFromServerType(site->server.GetType()));
+				if (site->GetOriginalServer().GetType() != DEFAULT) {
+					msg = wxString::Format(_("Remote path cannot be parsed. Make sure it is a valid absolute path and is supported by the current site's servertype (%s)."), CServer::GetNameFromServerType(site->GetOriginalServer().GetType()));
 				}
 				else {
 					msg = _("Remote path cannot be parsed. Make sure it is a valid absolute path.");
@@ -535,7 +539,7 @@ void CBookmarksDialog::UpdateBookmark()
 	wxString const remotePathRaw = xrc_call(*this, "ID_BOOKMARK_REMOTEDIR", &wxTextCtrl::GetValue);
 	if (!remotePathRaw.empty()) {
 		if (site) {
-			data->m_remote_dir.SetType(site->server.GetType());
+			data->m_remote_dir.SetType(site->GetOriginalServer().GetType());
 		}
 		data->m_remote_dir.SetPath(remotePathRaw.ToStdWstring());
 	}
@@ -602,7 +606,7 @@ void CBookmarksDialog::OnNewBookmark(wxCommandEvent&)
 
 		std::unique_ptr<Site> site;
 		if (!m_site_path.empty()) {
-			site = CSiteManager::GetSiteByPath(m_site_path).first;
+			site = CSiteManager::GetSiteByPath(options_, m_site_path).first;
 		}
 
 		if (!site) {
@@ -924,7 +928,7 @@ bool CBookmarksDialog::AddBookmark(const wxString &name, const wxString &local_d
 		AddTextElementUtf8(bookmark, "DirectoryComparison", "1");
 	}
 
-	if (!file.Save(false)) {
+	if (!file.Save()) {
 		wxString msg = wxString::Format(_("Could not write \"%s\", the bookmark could not be added: %s"), file.GetFileName(), file.GetError());
 		wxMessageBoxEx(msg, _("Error writing xml file"), wxICON_ERROR);
 		return false;
