@@ -160,7 +160,7 @@ void CEditHandler::RemoveTemporaryFilesInSpecificDir(std::wstring const& temp)
 			close(fd);
 		}
 #endif
-		fz::remove_file(fz::to_native(lockfile));
+		fz::remove_file(fz::to_native(lockfile), false);
 
 		if (wxFileName::FileExists(lockfile)) {
 			return;
@@ -341,7 +341,10 @@ int CEditHandler::GetFileCount(CEditHandler::fileType type, CEditHandler::fileSt
 
 bool CEditHandler::AddFile(CEditHandler::fileType type, std::wstring const& localFile, std::wstring const& remoteFile, CServerPath const& remotePath, Site const& site, int64_t size)
 {
-	wxASSERT(type != none);
+	if (type == none) {
+		wxBell();
+		return false;
+	}
 
 	fileState state;
 	if (type == local) {
@@ -1508,7 +1511,7 @@ int CEditHandlerStatusDialog::ShowModal()
 	GetSizer()->Fit(this);
 	SetMinClientSize(GetSizer()->GetMinSize());
 
-	impl_->windowStateManager_ = std::make_unique<CWindowStateManager>(static_cast<wxTopLevelWindow*>(this));
+	impl_->windowStateManager_ = std::make_unique<CWindowStateManager>(static_cast<wxTopLevelWindow*>(this), *COptions::Get());
 	impl_->windowStateManager_->Restore(OPTION_EDITSTATUSDIALOG_SIZE, GetSize());
 
 	SetCtrlState();
