@@ -1,31 +1,36 @@
-#include <filezilla.h>
+#include "filezilla.h"
 #include "sitemanager.h"
 
 #include "dialogex.h"
 #include "filezillaapp.h"
-#include "ipcmutex.h"
 #include "loginmanager.h"
 #include "Options.h"
 #include "xmlfunctions.h"
 
+#include "../commonui/fz_paths.h"
+#include "../commonui/ipcmutex.h"
+
 #include <libfilezilla/translate.hpp>
 
+#include <wx/menu.h>
+
 namespace {
 struct background_color {
-	wxColour const color;
+	site_colour const color;
 	char const*const name;
 };
 
+//note: the order needs to be the same as in site_colour (site.h) as that is used as index for this array
 background_color const background_colors[] = {
-	{ wxColour(), fztranslate_mark("None") },
-	{ wxColour(255, 0, 0, 32), fztranslate_mark("Red") },
-	{ wxColour(0, 255, 0, 32), fztranslate_mark("Green") },
-	{ wxColour(0, 0, 255, 32), fztranslate_mark("Blue") },
-	{ wxColour(255, 255, 0, 32), fztranslate_mark("Yellow") },
-	{ wxColour(0, 255, 255, 32), fztranslate_mark("Cyan") },
-	{ wxColour(255, 0, 255, 32), fztranslate_mark("Magenta") },
-	{ wxColour(255, 128, 0, 32), fztranslate_mark("Orange") },
-	{ wxColour(), 0 }
+	{ site_colour::none, fztranslate_mark("None") },
+	{ site_colour::red, fztranslate_mark("Red") },
+	{ site_colour::green, fztranslate_mark("Green") },
+	{ site_colour::blue, fztranslate_mark("Blue") },
+	{ site_colour::yellow, fztranslate_mark("Yellow") },
+	{ site_colour::cyan, fztranslate_mark("Cyan") },
+	{ site_colour::magenta, fztranslate_mark("Magenta") },
+	{ site_colour::orange, fztranslate_mark("Orange") },
+	{ site_colour::end_of_list, nullptr }
 };
 }
 
@@ -33,101 +38,12 @@ std::map<int, std::unique_ptr<Site>> CSiteManager::m_idMap;
 
 bool CSiteManager::Load(CSiteManagerXmlHandler& handler)
 {
-	CXmlFile file(wxGetApp().GetSettingsFile(_T("sitemanager")));
-	auto document = file.Load();
-	if (!document) {
-		wxMessageBoxEx(file.GetError(), _("Error loading xml file"), wxICON_ERROR);
-		return false;
+	std::wstring error;
+	bool ret = site_manager::Load(wxGetApp().GetSettingsFile(_T("sitemanager")), handler, error);
+	if (!ret) {
+		wxMessageBoxEx(error, _("Error loading xml file"), wxICON_ERROR);
 	}
-
-	auto element = document.child("Servers");
-	if (!element) {
-		return false;
-	}
-
-	return Load(element, handler);
-}
-
-bool CSiteManager::Load(pugi::xml_node element, CSiteManagerXmlHandler& handler)
-{
-	wxASSERT(element);
-
-	for (auto child = element.first_child(); child; child = child.next_sibling()) {
-		if (!strcmp(child.name(), "Folder")) {
-			std::wstring name = GetTextElement_Trimmed(child);
-			if (name.empty()) {
-				continue;
-			}
-
-			const bool expand = GetTextAttribute(child, "expanded") != _T("0");
-			if (!handler.AddFolder(name.substr(0, 255), expand)) {
-				return false;
-			}
-			Load(child, handler);
-			if (!handler.LevelUp()) {
-				return false;
-			}
-		}
-		else if (!strcmp(child.name(), "Server")) {
-			std::unique_ptr<Site> data = ReadServerElement(child);
-
-			if (data) {
-				handler.AddSite(std::move(data));
-			}
-		}
-	}
-
-	return true;
-}
-
-bool CSiteManager::ReadBookmarkElement(Bookmark & bookmark, pugi::xml_node element)
-{
-	bookmark.m_localDir = GetTextElement(element, "LocalDir");
-	bookmark.m_remoteDir.SetSafePath(GetTextElement(element, "RemoteDir"));
-
-	if (bookmark.m_localDir.empty() && bookmark.m_remoteDir.empty()) {
-		return false;
-	}
-
-	if (!bookmark.m_localDir.empty() && !bookmark.m_remoteDir.empty()) {
-		bookmark.m_sync = GetTextElementBool(element, "SyncBrowsing", false);
-	}
-
-	bookmark.m_comparison = GetTextElementBool(element, "DirectoryComparison", false);
-
-	return true;
-}
-
-std::unique_ptr<Site> CSiteManager::ReadServerElement(pugi::xml_node element)
-{
-	auto data = std::make_unique<Site>();
-	if (!::GetServer(element, *data)) {
-		return nullptr;
-	}
-	if (data->server.GetName().empty()) {
-		return nullptr;
-	}
-
-	data->comments_ = GetTextElement(element, "Comments");
-	data->m_colour = GetColourFromIndex(GetTextElementInt(element, "Colour"));
-
-	ReadBookmarkElement(data->m_default_bookmark, element);
-
-	// Bookmarks
-	for (auto bookmark = element.child("Bookmark"); bookmark; bookmark = bookmark.next_sibling("Bookmark")) {
-		std::wstring name = GetTextElement_Trimmed(bookmark, "Name");
-		if (name.empty()) {
-			continue;
-		}
-
-		Bookmark bookmarkData;
-		if (ReadBookmarkElement(bookmarkData, bookmark)) {
-			bookmarkData.m_name = name.substr(0, 255);
-			data->m_bookmarks.push_back(bookmarkData);
-		}
-	}
-
-	return data;
+	return ret;
 }
 
 class CSiteManagerXmlHandler_Menu : public CSiteManagerXmlHandler
@@ -182,12 +98,12 @@ public:
 
 	virtual bool AddSite(std::unique_ptr<Site> data)
 	{
-		std::wstring newName(data->server.GetName());
+		std::wstring newName(data->GetName());
 		int i = GetInsertIndex(m_pMenu, newName);
 		newName = LabelEscape(newName);
 		wxMenuItem* pItem = m_pMenu->Insert(i, wxID_ANY, newName);
 
-		data->SetSitePath(path + _T("/") + CSiteManager::EscapeSegment(data->server.GetName()));
+		data->SetSitePath(path + _T("/") + CSiteManager::EscapeSegment(data->GetName()));
 
 		(*m_idMap)[pItem->GetId()] = std::move(data);
 
@@ -306,29 +222,7 @@ void CSiteManager::ClearIdMap()
 
 bool CSiteManager::LoadPredefined(CSiteManagerXmlHandler& handler)
 {
-	CLocalPath const defaultsDir = wxGetApp().GetDefaultsDir();
-	if (defaultsDir.empty()) {
-		return false;
-	}
-
-	std::wstring const name(defaultsDir.GetPath() + _T("fzdefaults.xml"));
-	CXmlFile file(name);
-
-	auto document = file.Load();
-	if (!document) {
-		return false;
-	}
-
-	auto element = document.child("Servers");
-	if (!element) {
-		return false;
-	}
-
-	if (!Load(element, handler)) {
-		return false;
-	}
-
-	return true;
+	return site_manager::LoadPredefined(GetDefaultsDir(), handler);
 }
 
 std::unique_ptr<wxMenu> CSiteManager::GetSitesMenu_Predefined(std::map<int, std::unique_ptr<Site>> &idMap)
@@ -360,76 +254,13 @@ std::unique_ptr<Site> CSiteManager::GetSiteById(int id)
 	return pData;
 }
 
-bool CSiteManager::UnescapeSitePath(std::wstring path, std::vector<std::wstring>& result)
+std::pair<std::unique_ptr<Site>, Bookmark> CSiteManager::GetSiteByPath(COptionsBase & options, std::wstring const& sitePath, bool printErrors)
 {
-	result.clear();
+	std::wstring error;
 
-	std::wstring name;
-	wchar_t const* p = path.c_str();
-
-	// Undo escapement
-	bool lastBackslash = false;
-	while (*p) {
-		const wxChar& c = *p;
-		if (c == '\\') {
-			if (lastBackslash) {
-				name += _T("\\");
-				lastBackslash = false;
-			}
-			else {
-				lastBackslash = true;
-			}
-		}
-		else if (c == '/') {
-			if (lastBackslash) {
-				name += _T("/");
-				lastBackslash = 0;
-			}
-			else {
-				if (!name.empty()) {
-					result.push_back(name);
-				}
-				name.clear();
-			}
-		}
-		else {
-			name += *p;
-		}
-		++p;
-	}
-	if (lastBackslash) {
-		return false;
-	}
-	if (!name.empty()) {
-		result.push_back(name);
-	}
-
-	return !result.empty();
-}
-
-std::wstring CSiteManager::EscapeSegment(std::wstring segment)
-{
-	fz::replace_substrings(segment, _T("\\"), _T("\\\\"));
-	fz::replace_substrings(segment, _T("/"), _T("\\/"));
-	return segment;
-}
-
-std::wstring CSiteManager::BuildPath(wxChar root, std::vector<std::wstring> const& segments)
-{
-	std::wstring ret;
-	ret += root;
-	for (auto const& segment : segments) {
-		ret += _T("/") + EscapeSegment(segment);
-	}
-
-	return ret;
-}
-
-std::pair<std::unique_ptr<Site>, Bookmark> CSiteManager::GetSiteByPath(std::wstring const& sitePath, bool printErrors)
-{
-	wxString error;
-
-	auto ret = DoGetSiteByPath(sitePath, error);
+	CLocalPath settings_path{options.get_string(OPTION_DEFAULT_SETTINGSDIR)};
+	app_paths paths{settings_path, GetDefaultsDir()};
+	auto ret = site_manager::GetSiteByPath(paths, sitePath, error);
 	if (!ret.first && printErrors) {
 		wxMessageBoxEx(_("Site does not exist."), error);
 	}
@@ -437,86 +268,6 @@ std::pair<std::unique_ptr<Site>, Bookmark> CSiteManager::GetSiteByPath(std::wstr
 	return ret;
 }
 
-std::pair<std::unique_ptr<Site>, Bookmark> CSiteManager::DoGetSiteByPath(std::wstring sitePath, wxString& error)
-{
-	std::pair<std::unique_ptr<Site>, Bookmark> ret;
-	wxChar c = sitePath.empty() ? 0 : sitePath[0];
-	if (c != '0' && c != '1') {
-		error = _("Site path has to begin with 0 or 1.");
-		return ret;
-	}
-
-	sitePath = sitePath.substr(1);
-
-	// We have to synchronize access to sitemanager.xml so that multiple processed don't write
-	// to the same file or one is reading while the other one writes.
-	CInterProcessMutex mutex(MUTEX_SITEMANAGER);
-
-	CXmlFile file;
-	if (c == '0') {
-		file.SetFileName(wxGetApp().GetSettingsFile(_T("sitemanager")));
-	}
-	else {
-		CLocalPath const defaultsDir = wxGetApp().GetDefaultsDir();
-		if (defaultsDir.empty()) {
-			error = _("Site does not exist.");
-			return ret;
-		}
-		file.SetFileName(defaultsDir.GetPath() + _T("fzdefaults.xml"));
-	}
-
-	auto document = file.Load();
-	if (!document) {
-		wxMessageBoxEx(file.GetError(), _("Error loading xml file"), wxICON_ERROR);
-		return ret;
-	}
-
-	auto element = document.child("Servers");
-	if (!element) {
-		error = _("Site does not exist.");
-		return ret;
-	}
-
-	std::vector<std::wstring> segments;
-	if (!UnescapeSitePath(sitePath, segments) || segments.empty()) {
-		error = _("Site path is malformed.");
-		return ret;
-	}
-
-	auto child = GetElementByPath(element, segments);
-	if (!child) {
-		error = _("Site does not exist.");
-		return ret;
-	}
-
-	pugi::xml_node bookmark;
-	if (!strcmp(child.name(), "Bookmark")) {
-		bookmark = child;
-		child = child.parent();
-		segments.pop_back();
-	}
-
-	ret.first = ReadServerElement(child);
-	if (!ret.first) {
-		error = _("Could not read server item.");
-	}
-	else {
-		if (bookmark) {
-			Bookmark bm;
-			if (ReadBookmarkElement(bm, bookmark)) {
-				ret.second = bm;
-			}
-		}
-		else {
-			ret.second = ret.first->m_default_bookmark;
-		}
-
-		ret.first->SetSitePath(BuildPath(c, segments));
-	}
-
-	return ret;
-}
-
 std::wstring CSiteManager::AddServer(Site site)
 {
 	// We have to synchronize access to sitemanager.xml so that multiple processed don't write
@@ -564,14 +315,14 @@ std::wstring CSiteManager::AddServer(Site site)
 		name = _("New site").ToStdWstring() + fz::sprintf(L" %d", ++i);
 	}
 
-	site.server.SetName(name);
+	site.SetName(name);
 
 	auto xServer = element.append_child("Server");
 	SetServer(xServer, site);
 	AddTextElement(xServer, name);
 
-	if (!file.Save(false)) {
-		if (COptions::Get()->GetOptionVal(OPTION_DEFAULT_KIOSKMODE) == 2) {
+	if (!file.Save()) {
+		if (COptions::Get()->get_int(OPTION_DEFAULT_KIOSKMODE) == 2) {
 			return std::wstring();
 		}
 
@@ -583,39 +334,7 @@ std::wstring CSiteManager::AddServer(Site site)
 	return L"0/" + EscapeSegment(name);
 }
 
-pugi::xml_node CSiteManager::GetElementByPath(pugi::xml_node node, std::vector<std::wstring> const& segments)
-{
-	for (auto const& segment : segments) {
-		pugi::xml_node child;
-		for (child = node.first_child(); child; child = child.next_sibling()) {
-			if (strcmp(child.name(), "Server") && strcmp(child.name(), "Folder") && strcmp(child.name(), "Bookmark")) {
-				continue;
-			}
-
-			std::wstring name = GetTextElement_Trimmed(child, "Name");
-			if (name.empty()) {
-				name = GetTextElement_Trimmed(child);
-			}
-			if (name.empty()) {
-				continue;
-			}
-
-			if (name == segment) {
-				break;
-			}
-		}
-		if (!child) {
-			return pugi::xml_node();
-		}
-
-		node = child;
-		continue;
-	}
-
-	return node;
-}
-
-bool CSiteManager::AddBookmark(std::wstring sitePath, const wxString& name, const wxString &local_dir, const CServerPath &remote_dir, bool sync, bool comparison)
+bool CSiteManager::AddBookmark(std::wstring sitePath, wxString const& name, wxString const& local_dir, CServerPath const& remote_dir, bool sync, bool comparison)
 {
 	if (local_dir.empty() && remote_dir.empty()) {
 		return false;
@@ -695,8 +414,8 @@ bool CSiteManager::AddBookmark(std::wstring sitePath, const wxString& name, cons
 		AddTextElementUtf8(bookmark, "DirectoryComparison", "1");
 	}
 
-	if (!file.Save(false)) {
-		if (COptions::Get()->GetOptionVal(OPTION_DEFAULT_KIOSKMODE) == 2) {
+	if (!file.Save()) {
+		if (COptions::Get()->get_int(OPTION_DEFAULT_KIOSKMODE) == 2) {
 			return true;
 		}
 
@@ -752,8 +471,8 @@ bool CSiteManager::ClearBookmarks(std::wstring sitePath)
 		bookmark = child.child("Bookmark");
 	}
 
-	if (!file.Save(false)) {
-		if (COptions::Get()->GetOptionVal(OPTION_DEFAULT_KIOSKMODE) == 2) {
+	if (!file.Save()) {
+		if (COptions::Get()->get_int(OPTION_DEFAULT_KIOSKMODE) == 2) {
 			return true;
 		}
 
@@ -772,15 +491,7 @@ bool CSiteManager::HasSites()
 	return handler.sites_ > 0;
 }
 
-wxColour CSiteManager::GetColourFromIndex(int i)
-{
-	if (i < 0 || static_cast<unsigned int>(i) + 1 >= (sizeof(background_colors) / sizeof(background_color) + 1)) {
-		return wxColour();
-	}
-	return background_colors[i].color;
-}
-
-int CSiteManager::GetColourIndex(wxColour const& c)
+int CSiteManager::GetColourIndex(site_colour const& c)
 {
 	for (int i = 0; background_colors[i].name; ++i) {
 		if (c == background_colors[i].color) {
@@ -799,30 +510,30 @@ wxString CSiteManager::GetColourName(int i)
 	return wxGetTranslation(background_colors[i].name);
 }
 
-void CSiteManager::Rewrite(CLoginManager & loginManager, pugi::xml_node element, bool on_failure_set_to_ask)
+void CSiteManager::Rewrite(CLoginManager & loginManager, COptionsBase& options, pugi::xml_node element, bool on_failure_set_to_ask)
 {
-	bool const forget = COptions::Get()->GetOptionVal(OPTION_DEFAULT_KIOSKMODE) != 0;
+	bool const forget = options.get_int(OPTION_DEFAULT_KIOSKMODE) != 0;
 	for (auto child = element.first_child(); child; child = child.next_sibling()) {
 		if (!strcmp(child.name(), "Folder")) {
-			Rewrite(loginManager, child, on_failure_set_to_ask);
+			Rewrite(loginManager, options, child, on_failure_set_to_ask);
 		}
 		else if (!strcmp(child.name(), "Server")) {
 			auto site = ReadServerElement(child);
 			if (site) {
 				if (!forget) {
 					loginManager.AskDecryptor(site->credentials.encrypted_, true, false);
-					site->credentials.Unprotect(loginManager.GetDecryptor(site->credentials.encrypted_), on_failure_set_to_ask);
+					unprotect(site->credentials, loginManager.GetDecryptor(site->credentials.encrypted_), on_failure_set_to_ask);
 				}
-				site->credentials.Protect();
-				Save(child, *site);
+				protect(site->credentials);
+				site_manager::Save(child, *site, loginManager, options);
 			}
 		}
 	}
 }
 
-void CSiteManager::Rewrite(CLoginManager & loginManager, bool on_failure_set_to_ask)
+void CSiteManager::Rewrite(CLoginManager & loginManager, COptionsBase& options, bool on_failure_set_to_ask)
 {
-	if (COptions::Get()->GetOptionVal(OPTION_DEFAULT_KIOSKMODE) == 2) {
+	if (options.get_int(OPTION_DEFAULT_KIOSKMODE) == 2) {
 		return;
 	}
 	CInterProcessMutex mutex(MUTEX_SITEMANAGER);
@@ -839,44 +550,9 @@ void CSiteManager::Rewrite(CLoginManager & loginManager, bool on_failure_set_to_
 		return;
 	}
 
-	Rewrite(loginManager, element, on_failure_set_to_ask);
+	Rewrite(loginManager, options, element, on_failure_set_to_ask);
 
-	file.Save(true);
-}
-
-void CSiteManager::Save(pugi::xml_node element, Site const& site)
-{
-	SetServer(element, site);
-
-	// Save comments
-	AddTextElement(element, "Comments", site.comments_);
-
-	// Save colour
-	AddTextElement(element, "Colour", CSiteManager::GetColourIndex(site.m_colour));
-
-	// Save local dir
-	AddTextElement(element, "LocalDir", site.m_default_bookmark.m_localDir);
-
-	// Save remote dir
-	AddTextElement(element, "RemoteDir", site.m_default_bookmark.m_remoteDir.GetSafePath());
-
-	AddTextElementUtf8(element, "SyncBrowsing", site.m_default_bookmark.m_sync ? "1" : "0");
-	AddTextElementUtf8(element, "DirectoryComparison", site.m_default_bookmark.m_comparison ? "1" : "0");
-
-	for (auto const& bookmark : site.m_bookmarks) {
-		auto node = element.append_child("Bookmark");
-
-		AddTextElement(node, "Name", bookmark.m_name);
-
-		// Save local dir
-		AddTextElement(node, "LocalDir", bookmark.m_localDir);
-
-		// Save remote dir
-		AddTextElement(node, "RemoteDir", bookmark.m_remoteDir.GetSafePath());
-
-		AddTextElementUtf8(node, "SyncBrowsing", bookmark.m_sync ? "1" : "0");
-		AddTextElementUtf8(node, "DirectoryComparison", bookmark.m_comparison ? "1" : "0");
-	}
+	SaveWithErrorDialog(file);
 }
 
 namespace {
@@ -892,7 +568,7 @@ pugi::xml_node GetChildWithName(pugi::xml_node element, std::wstring const& name
 			return child;
 		}
 	}
-	
+
 	return pugi::xml_node();
 }
 
@@ -935,7 +611,7 @@ bool CSiteManager::ImportSites(pugi::xml_node sites)
 		return false;
 	}
 
-	return file.Save(true);
+	return SaveWithErrorDialog(file);
 }
 
 bool CSiteManager::ImportSites(pugi::xml_node sitesToImport, pugi::xml_node existingSites)
@@ -966,21 +642,21 @@ bool CSiteManager::ImportSites(pugi::xml_node sitesToImport, pugi::xml_node exis
 		if (!site) {
 			continue;
 		}
-			
+
 		// Find free name
-		auto const name = site->server.GetName();
+		auto const& name = site->GetName();
 		std::wstring newName = name;
 		int i = 2;
 		while (GetChildWithName(existingSites, newName)) {
-			newName = fz::sprintf(L"%s %d", name.substr(0, 240), i++);
+			newName = fz::sprintf(L"%s %d", std::wstring_view(name).substr(0, 240), i++);
 		}
-		site->server.SetName(newName);
+		site->SetName(newName);
 
-		site->credentials.Unprotect(loginManager.GetDecryptor(site->credentials.encrypted_), false);
-		site->credentials.Protect();
+		unprotect(site->credentials, loginManager.GetDecryptor(site->credentials.encrypted_), false);
+		protect(site->credentials);
 
 		auto xsite = existingSites.append_child("Server");
-		Save(xsite, *site);
+		site_manager::Save(xsite, *site, loginManager, *COptions::Get());
 	}
 
 	return true;
