@@ -408,7 +408,7 @@ bool CSiteManagerDialog::Create(wxWindow* parent, std::vector<_connected_site>*
 	}
 	SetCtrlState();
 
-	m_pWindowStateManager = new CWindowStateManager(this);
+	m_pWindowStateManager = new CWindowStateManager(this, options_);
 	m_pWindowStateManager->Restore(OPTION_SITEMANAGER_POSITION);
 
 	tree_->SetDropTarget(new CSiteManagerDropTarget(this));
@@ -872,6 +872,7 @@ void CSiteManagerDialog::OnNewFolder(wxCommandEvent&)
 	if (!Verify()) {
 		return;
 	}
+	UpdateItem();
 
 	wxString name = FindFirstFreeName(item, _("New folder"));
 
@@ -1127,7 +1128,7 @@ void CSiteManagerDialog::OnSelChanged(wxTreeEvent& evt)
 		m_is_deleting = false;
 	}
 
-#if !defined(__WXMSW__) && wxCHECK_VERSION(3, 2, 1)
+#if !defined(__WXMSW__)
 	auto focused = tree_->GetFocusedItem();
 	if (focused) {
 		tree_->EnsureVisible(focused);
@@ -1157,6 +1158,7 @@ void CSiteManagerDialog::OnNewSite(wxCommandEvent&)
 	if (!Verify()) {
 		return;
 	}
+	UpdateItem();
 
 	Site site;
 	site.server.SetProtocol(ServerProtocol::FTP);
